//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/compat.rs: Compatibility code for different libcs
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
// waitid and WaitStatus are based in part upon nix which is:
//     Copyright (c) nix authors
//     SPDX-License-Identifier: MIT
//
// SPDX-License-Identifier: GPL-3.0-or-later

#![allow(non_camel_case_types)]

use std::{
    ffi::{OsStr, OsString},
    os::{fd::RawFd, unix::ffi::OsStrExt},
    path::{Path, PathBuf},
};

use nix::{
    errno::Errno,
    sys::{
        socket::SockaddrLike,
        wait::{Id, WaitPidFlag, WaitStatus as NixWaitStatus},
    },
    unistd::Pid,
};
use nonempty::NonEmpty;

/// Aligns the given length to the nearest 4-byte boundary.
///
/// This function is useful for ensuring that data structures in a 32-bit architecture
/// are correctly aligned to 4 bytes as required by many system interfaces for proper operation.
///
/// # Arguments
///
/// * `len` - The length that needs to be aligned.
///
/// # Returns
///
/// The smallest length that is a multiple of 4 bytes and is not less than `len`.
pub(crate) const fn cmsg_align_32(len: usize) -> usize {
    len.saturating_add(3) & !3
}

/// Calculates the total space required for a control message including data and padding on a 32-bit system.
///
/// This function is specifically tailored for 32-bit architectures where control message
/// headers and data need to be aligned to 4-byte boundaries.
///
/// # Arguments
///
/// * `length` - The length of the data in the control message.
///
/// # Returns
///
/// The total space in bytes required to store the control message, ensuring proper alignment.
pub(crate) const fn cmsg_space_32(length: u32) -> usize {
    cmsg_align_32((length as usize).saturating_add(cmsg_align_32(std::mem::size_of::<cmsghdr32>())))
}

/// Computes the byte length of a control message's header and data for a 32-bit system.
///
/// This function helps in determining the correct length for control messages where the
/// header and data must be aligned to 4-byte boundaries in a 32-bit architecture.
///
/// # Arguments
///
/// * `length` - The length of the data part of the control message.
///
/// # Returns
///
/// The combined length of the control message header and the data, aligned as required.
pub(crate) const fn cmsg_len_32(length: u32) -> usize {
    cmsg_align_32(std::mem::size_of::<cmsghdr32>()).saturating_add(length as usize)
}

/// This structure represents the Linux 32-bit data structure 'struct iovec'
#[repr(C)]
pub struct iovec32 {
    iov_base: u32,
    iov_len: u32,
}

impl From<iovec32> for nix::libc::iovec {
    fn from(src: iovec32) -> Self {
        nix::libc::iovec {
            iov_base: src.iov_base as *mut _,
            iov_len: src.iov_len as usize,
        }
    }
}

/// This structure represents the Linux 32-bit data structure 'struct mmmsghdr'
#[derive(Copy, Clone)]
#[repr(C)]
pub struct mmsghdr32 {
    pub msg_hdr: msghdr32,
    pub msg_len: u32,
}

/// This structure represents the Linux native data structure 'struct mmsghdr'
#[derive(Copy, Clone)]
#[repr(C)]
pub struct mmsghdr {
    pub msg_hdr: msghdr,
    pub msg_len: nix::libc::c_uint,
}

impl From<mmsghdr32> for mmsghdr {
    fn from(src: mmsghdr32) -> Self {
        mmsghdr {
            msg_hdr: msghdr::from(src.msg_hdr),
            msg_len: src.msg_len,
        }
    }
}

impl From<mmsghdr> for mmsghdr32 {
    fn from(src: mmsghdr) -> Self {
        mmsghdr32 {
            msg_hdr: msghdr32::from(src.msg_hdr),
            msg_len: src.msg_len,
        }
    }
}

/// This structure represents the Linux 32-bit data structure 'struct cmsghdr'
#[repr(C)]
pub struct cmsghdr32 {
    pub cmsg_len: u32,
    pub cmsg_level: i32,
    pub cmsg_type: i32,
}

/// This structure represents the Linux native data structure 'struct cmsghdr'
#[repr(C)]
pub struct cmsghdr {
    pub cmsg_len: nix::libc::size_t,
    pub cmsg_level: nix::libc::c_int,
    pub cmsg_type: nix::libc::c_int,
}

impl From<cmsghdr32> for cmsghdr {
    fn from(src: cmsghdr32) -> Self {
        cmsghdr {
            cmsg_len: src.cmsg_len as nix::libc::size_t,
            cmsg_level: src.cmsg_level,
            cmsg_type: src.cmsg_type,
        }
    }
}

/// This structure represents the Linux 32-bit data structure 'struct msghdr'
#[derive(Copy, Clone)]
#[repr(C)]
pub struct msghdr32 {
    // In 32-bit systems, pointers are 32-bit.
    pub msg_name: u32,       // Use u32 to represent a 32-bit pointer.
    pub msg_namelen: u32,    // socklen_t is typically 32-bit.
    pub msg_iov: u32,        // Use u32 to represent a 32-bit pointer to iovec.
    pub msg_iovlen: u32,     // size_t is 32-bit on 32-bit systems.
    pub msg_control: u32,    // Use u32 to represent a 32-bit pointer.
    pub msg_controllen: u32, // size_t is 32-bit on 32-bit systems.
    pub msg_flags: i32,      // c_int remains the same (32-bit).
}

/// This structure represents the Linux native data structure 'struct msghdr'
#[derive(Copy, Clone)]
#[repr(C)]
pub struct msghdr {
    pub msg_name: *mut nix::libc::c_void,
    pub msg_namelen: nix::libc::socklen_t,
    pub msg_iov: *mut nix::libc::iovec,
    pub msg_iovlen: nix::libc::size_t,
    pub msg_control: *mut nix::libc::c_void,
    pub msg_controllen: nix::libc::size_t,
    pub msg_flags: nix::libc::c_int,
}

impl From<msghdr32> for msghdr {
    fn from(msg: msghdr32) -> Self {
        msghdr {
            msg_name: msg.msg_name as *mut nix::libc::c_void,
            msg_namelen: msg.msg_namelen as nix::libc::socklen_t,
            msg_iov: msg.msg_iov as *mut nix::libc::iovec,
            msg_iovlen: msg.msg_iovlen as nix::libc::size_t,
            msg_control: msg.msg_control as *mut nix::libc::c_void,
            msg_controllen: msg.msg_controllen as nix::libc::size_t,
            msg_flags: msg.msg_flags as nix::libc::c_int,
        }
    }
}

#[allow(clippy::unnecessary_cast)]
#[allow(clippy::cast_possible_truncation)]
impl From<msghdr> for msghdr32 {
    fn from(msg: msghdr) -> Self {
        msghdr32 {
            msg_name: msg.msg_name as u32,
            msg_namelen: msg.msg_namelen as u32,
            msg_iov: msg.msg_iov as u32,
            msg_iovlen: msg.msg_iovlen as u32,
            msg_control: msg.msg_control as u32,
            msg_controllen: msg.msg_controllen as u32,
            msg_flags: msg.msg_flags as i32,
        }
    }
}

/// This structure represents the Linux data structure `struct statx_timestamp`
#[repr(C)]
pub(crate) struct statx_timestamp {
    tv_sec: i64,
    tv_nsec: u32,
    __statx_timestamp_pad1: [i32; 1],
}

/// This structure represents the Linux data structure `struct statx`
#[repr(C)]
pub(crate) struct statx {
    stx_mask: u32,       // What results were written [uncond]
    stx_blksize: u32,    // Preferred general I/O size [uncond]
    stx_attributes: u64, // Flags conveying information about the file [uncond]

    stx_nlink: u32, // Number of hard links
    stx_uid: u32,   // User ID of owner
    stx_gid: u32,   // Group ID of owner
    stx_mode: u16,  // File mode
    __statx_pad1: [u16; 1],

    stx_ino: u64,             // Inode number
    stx_size: u64,            // File size
    stx_blocks: u64,          // Number of 512-byte blocks allocated
    stx_attributes_mask: u64, // Mask to show what's supported in stx_attributes

    stx_atime: statx_timestamp, // Last access time
    stx_btime: statx_timestamp, // File creation time
    stx_ctime: statx_timestamp, // Last attribute change time
    stx_mtime: statx_timestamp, // Last data modification time

    stx_rdev_major: u32, // Device ID of special file [if bdev/cdev]
    stx_rdev_minor: u32,
    stx_dev_major: u32, // ID of device containing file [uncond]
    stx_dev_minor: u32,

    stx_mnt_id: u64,
    stx_dio_mem_align: u32,    // Memory buffer alignment for direct I/O
    stx_dio_offset_align: u32, // File offset alignment for direct I/O

    __statx_pad2: [u64; 12], // Spare space for future expansion
}

/// This struct represents a directory entry.
#[derive(Clone)]
pub(crate) struct DirEntry {
    /// The `dirent64` structure.
    pub(crate) dirent: Box<[u8]>,

    /// Size of the file name, in bytes.
    pub(crate) namelen: usize,
}

impl std::fmt::Debug for DirEntry {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_tuple("DirEntry").field(&self.name()).finish()
    }
}

impl DirEntry {
    /// Return a `PathBuf` representing the name of the directory entry.
    pub fn to_path_buf(&self) -> PathBuf {
        Path::new(&self.name()).to_path_buf()
    }

    /// Return the name of the directory entry.
    pub fn name(&self) -> OsString {
        // SAFETY: We trust self.dirent points to a valid `dirent64` structure.
        OsStr::from_bytes(unsafe { self.name_bytes() }).to_os_string()
    }

    unsafe fn name_bytes(&self) -> &[u8] {
        let dirent: *const nix::libc::dirent64 = Box::as_ref(&self.dirent).as_ptr().cast();
        let d_name: *const nix::libc::c_char = (*dirent).d_name.as_ptr();
        std::slice::from_raw_parts(d_name.cast(), self.namelen)
    }
}

/// Represents a list of directory entries as returned by `getdents64` system call.
pub(crate) type Directory = NonEmpty<DirEntry>;

/// Retrieve directory entries from an open directory file descriptor.
///
/// # Parameters
///
/// - `fd`: The open directory file descriptor.
///
/// # Returns
///
/// `Some(Directory)` containing the directory entries,
/// `None` when there are no more entries.
///
/// # Safety
///
/// This function calls the `getdents64` system call directly which is an unsafe function.
/// Ensure that `fd` is a valid open directory file descriptor to avoid undefined behavior.
pub(crate) fn getdents(fd: RawFd, count: usize) -> Result<Option<Directory>, Errno> {
    let mut buffer = vec![0u8; count];

    let size = loop {
        // SAFETY: In libc, we trust.
        match unsafe { getdents64(fd, buffer.as_mut_ptr().cast(), buffer.len()) } {
            0 => {
                return Ok(None);
            }
            error if error < 0 => {
                #[allow(clippy::arithmetic_side_effects)]
                #[allow(clippy::cast_possible_truncation)]
                let error = -error as i32;
                if error == Errno::EINTR as i32 {
                    continue;
                }
                return Err(Errno::from_i32(error));
            }
            size =>
            {
                #[allow(clippy::cast_possible_truncation)]
                #[allow(clippy::cast_sign_loss)]
                break size as usize
            }
        }
    };

    // Parsing the buffer to extract DirEntry structures
    let mut entries = Vec::new();
    let mut offset = 0;
    while offset < size {
        // SAFETY: In libc, we trust.
        unsafe {
            let dirent: *const nix::libc::dirent64 = buffer.as_ptr().add(offset).cast();
            let d_reclen = (*dirent).d_reclen as usize;

            // Copy the bytes of the `dirent64` structure from the buffer to a
            // memory owned by `DirEntry`.
            let entry = std::slice::from_raw_parts(dirent.cast(), d_reclen)
                .to_owned()
                .into_boxed_slice();

            entries.push(DirEntry {
                dirent: entry,
                namelen: nix::libc::strlen((*dirent).d_name.as_ptr()),
            });

            offset = offset.saturating_add(d_reclen);
        }
    }

    if entries.is_empty() {
        Ok(None)
    } else {
        #[allow(clippy::disallowed_methods)]
        Ok(Some(NonEmpty::from_vec(entries).unwrap()))
    }
}

/// Wrapper for the `getdents64` syscall
#[inline(always)]
unsafe fn getdents64(
    fd: nix::libc::c_int,
    buf: *mut nix::libc::c_void,
    bytes: nix::libc::size_t,
) -> nix::libc::c_long {
    nix::libc::syscall(nix::libc::SYS_getdents64, fd, buf, bytes)
}

/// WaitStatus with support for signals that nix' Signal type don't support.
#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
pub(crate) enum WaitStatus {
    Exited(Pid, i32),
    Signaled(Pid, i32, bool),
    Stopped(Pid, i32),
    PtraceEvent(Pid, i32, nix::libc::c_int),
    PtraceSyscall(Pid),
    Continued(Pid),
    StillAlive,
}

impl From<NixWaitStatus> for WaitStatus {
    fn from(status: NixWaitStatus) -> Self {
        match status {
            NixWaitStatus::Exited(pid, code) => WaitStatus::Exited(pid, code),
            NixWaitStatus::Signaled(pid, signal, core_dump) => {
                WaitStatus::Signaled(pid, signal as i32, core_dump)
            }
            NixWaitStatus::Stopped(pid, signal) => WaitStatus::Stopped(pid, signal as i32),
            NixWaitStatus::PtraceEvent(pid, signal, event) => {
                WaitStatus::PtraceEvent(pid, signal as i32, event)
            }
            NixWaitStatus::PtraceSyscall(pid) => WaitStatus::PtraceSyscall(pid),
            NixWaitStatus::Continued(pid) => WaitStatus::Continued(pid),
            NixWaitStatus::StillAlive => WaitStatus::StillAlive,
        }
    }
}

fn exited(status: i32) -> bool {
    nix::libc::WIFEXITED(status)
}

fn exit_status(status: i32) -> i32 {
    nix::libc::WEXITSTATUS(status)
}

fn signaled(status: i32) -> bool {
    nix::libc::WIFSIGNALED(status)
}

fn term_signal(status: i32) -> i32 {
    nix::libc::WTERMSIG(status)
}

fn dumped_core(status: i32) -> bool {
    nix::libc::WCOREDUMP(status)
}

fn stopped(status: i32) -> bool {
    nix::libc::WIFSTOPPED(status)
}

fn stop_signal(status: i32) -> i32 {
    nix::libc::WSTOPSIG(status)
}

fn syscall_stop(status: i32) -> bool {
    // From ptrace(2), setting PTRACE_O_TRACESYSGOOD has the effect
    // of delivering SIGTRAP | 0x80 as the signal number for syscall
    // stops. This allows easily distinguishing syscall stops from
    // genuine SIGTRAP signals.
    nix::libc::WSTOPSIG(status) == libc::SIGTRAP | 0x80
}

fn stop_additional(status: i32) -> nix::libc::c_int {
    (status >> 16) as nix::libc::c_int
}

fn continued(status: i32) -> bool {
    nix::libc::WIFCONTINUED(status)
}

impl WaitStatus {
    pub(crate) fn from_raw(pid: Pid, status: i32) -> WaitStatus {
        if exited(status) {
            WaitStatus::Exited(pid, exit_status(status))
        } else if signaled(status) {
            WaitStatus::Signaled(pid, term_signal(status), dumped_core(status))
        } else if stopped(status) {
            let status_additional = stop_additional(status);
            if syscall_stop(status) {
                WaitStatus::PtraceSyscall(pid)
            } else if status_additional == 0 {
                WaitStatus::Stopped(pid, stop_signal(status))
            } else {
                WaitStatus::PtraceEvent(pid, stop_signal(status), stop_additional(status))
            }
        } else {
            assert!(continued(status));
            WaitStatus::Continued(pid)
        }
    }
}

/// Wrapper for the `waitid` syscall
/// This is identical to nix' waitid except we use our custom WaitStatus.
pub(crate) fn waitid(id: Id, flags: WaitPidFlag) -> Result<WaitStatus, Errno> {
    #[allow(clippy::cast_sign_loss)]
    let (idtype, idval) = match id {
        Id::All => (nix::libc::P_ALL, 0),
        Id::Pid(pid) => (nix::libc::P_PID, pid.as_raw() as nix::libc::id_t),
        Id::PGid(pid) => (nix::libc::P_PGID, pid.as_raw() as nix::libc::id_t),
        Id::PIDFd(fd) => (nix::libc::P_PIDFD, fd as nix::libc::id_t),
    };

    // SAFETY: In libc, we trust.
    let siginfo = unsafe {
        // Memory is zeroed rather than uninitialized, as not all platforms
        // initialize the memory in the StillAlive case
        let mut siginfo: libc::siginfo_t = std::mem::zeroed();
        Errno::result(nix::libc::waitid(idtype, idval, &mut siginfo, flags.bits()))?;
        siginfo
    };

    // SAFETY: In libc, we trust.
    let si_pid = unsafe { siginfo.si_pid() };
    if si_pid == 0 {
        return Ok(WaitStatus::StillAlive);
    }

    assert_eq!(siginfo.si_signo, nix::libc::SIGCHLD);

    let pid = Pid::from_raw(si_pid);
    // SAFETY: In libc, we trust.
    let si_status = unsafe { siginfo.si_status() };

    let status = match siginfo.si_code {
        nix::libc::CLD_EXITED => WaitStatus::Exited(pid, si_status),
        nix::libc::CLD_KILLED | nix::libc::CLD_DUMPED => {
            WaitStatus::Signaled(pid, si_status, siginfo.si_code == nix::libc::CLD_DUMPED)
        }
        nix::libc::CLD_STOPPED => WaitStatus::Stopped(pid, si_status),
        nix::libc::CLD_CONTINUED => WaitStatus::Continued(pid),
        nix::libc::CLD_TRAPPED => {
            if si_status == nix::libc::SIGTRAP | 0x80 {
                WaitStatus::PtraceSyscall(pid)
            } else {
                WaitStatus::PtraceEvent(pid, si_status & 0xff, (si_status >> 8) as nix::libc::c_int)
            }
        }
        _ => return Err(Errno::EINVAL),
    };

    Ok(status)
}

#[allow(clippy::cast_possible_truncation)]
pub(crate) const PF_UNIX: nix::libc::sa_family_t = nix::libc::AF_UNIX as nix::libc::sa_family_t;
#[allow(clippy::cast_possible_truncation)]
pub(crate) const PF_INET: nix::libc::sa_family_t = nix::libc::AF_INET as nix::libc::sa_family_t;
#[allow(clippy::cast_possible_truncation)]
pub(crate) const PF_INET6: nix::libc::sa_family_t = nix::libc::AF_INET6 as nix::libc::sa_family_t;
#[allow(clippy::cast_possible_truncation)]
pub(crate) const PF_ALG: nix::libc::sa_family_t = nix::libc::AF_ALG as nix::libc::sa_family_t;

/// nix' SockAddrLike.family() function does not support AF_ALG.
/// This is a workaround that accesses the underlying sockaddr directly.
pub(crate) fn addr_family<T: SockaddrLike>(addr: &T) -> nix::libc::sa_family_t {
    // SAFETY: This is safe as long as addr.as_ptr() returns a valid pointer to a sockaddr.
    unsafe { (*addr.as_ptr()).sa_family }
}
