//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/fs.rs: Filesystem utilities
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
// Based in part upon uutils coreutils package's src/lib/features/fs.rs which is:
//   (c) Joseph Crail <jbcrail@gmail.com>
//   (c) Jian Zeng <anonymousknight96 AT gmail.com>
// Tests base based in part upon gnulib packages' tests/test-canonicalize.c which is:
//   (c) Free Software Foundation, Inc.
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! Set of functions to manage files and symlinks

use std::{
    collections::{HashSet, VecDeque},
    ffi::{CStr, OsString},
    fs::{metadata, set_permissions, File},
    hash::Hash,
    io::{Read, Write},
    os::{
        fd::{AsRawFd, FromRawFd, OwnedFd, RawFd},
        unix::{
            ffi::{OsStrExt, OsStringExt},
            fs::PermissionsExt,
        },
    },
    path::{Component, Path, PathBuf},
};

use nix::{
    errno::Errno,
    fcntl::{fcntl, open, readlink, FcntlArg, OFlag},
    sys::stat::{fstat, lstat, stat, Mode, SFlag},
    unistd::{access, AccessFlags, Pid},
    NixPath,
};
use openat2::{openat2, OpenHow, ResolveFlags};
use serde::{Serialize, Serializer};
use smallvec::SmallVec;

use crate::proc::{proc_cwd, proc_fd, proc_tgid};

/// Information to uniquely identify a file
pub struct FileInformation(pub nix::sys::stat::FileStat);

impl FileInformation {
    /// Get information for a given link.
    pub fn from_link<P: AsRef<Path>>(path: P) -> Result<Self, Errno> {
        lstat(path.as_ref()).map(Self)
    }

    /// Get information from a given path.
    pub fn from_path<P: AsRef<Path>>(path: P) -> Result<Self, Errno> {
        stat(path.as_ref()).map(Self)
    }
}

impl PartialEq for FileInformation {
    fn eq(&self, other: &Self) -> bool {
        self.0.st_dev == other.0.st_dev && self.0.st_ino == other.0.st_ino
    }
}

impl Eq for FileInformation {}

impl Hash for FileInformation {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.0.st_dev.hash(state);
        self.0.st_ino.hash(state);
    }
}

/// Controls how missing components should be handled when canonicalizing a path.
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum MissingHandling {
    /// Last component may exist, other components must exist.
    Normal,

    /// All components must exist.
    Existing,

    /// Last component must not exist, other componenets must exist.
    Missing,
}

impl Default for MissingHandling {
    fn default() -> Self {
        Self::Normal
    }
}

impl Serialize for MissingHandling {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let variant_str = match *self {
            MissingHandling::Normal => "normal",
            MissingHandling::Existing => "exist",
            MissingHandling::Missing => "miss",
        };

        serializer.serialize_str(variant_str)
    }
}

/// Returns true if the given `Path` ends with a slash.
#[inline]
pub fn path_ends_with_slash<P: AsRef<Path>>(path: P) -> bool {
    path.as_ref().as_os_str().as_bytes().last() == Some(&b'/')
}

/// Resolve a symbolic link honouring magic proc links.
pub fn resolve_symlink<P: AsRef<Path>>(pid: Pid, path: P) -> Result<PathBuf, Errno> {
    // SAFETY: Handle /proc/self and /proc/thread-self magic links
    match path.as_ref().as_os_str().as_bytes() {
        b"/proc/self" => {
            let mut b = itoa::Buffer::new();
            Ok(PathBuf::from(b.format(pid.as_raw())))
        }
        b"/proc/thread-self" => {
            let mut b0 = itoa::Buffer::new();
            let mut b1 = itoa::Buffer::new();
            let mut p = PathBuf::from(b0.format(proc_tgid(pid)?.as_raw()));
            p.push("task");
            p.push(b1.format(pid.as_raw()));
            Ok(p)
        }
        _ => read_link(&path),
    }
}

/// Resolve a symbolic link.
pub fn read_link<P: AsRef<Path>>(path: P) -> Result<PathBuf, Errno> {
    readlink(path.as_ref()).map(PathBuf::from)
}

/// Check if a path exists.
/// This is more efficient than std::path::Path.exists() because
/// it uses access() rather than stat().
pub fn path_exists<P: AsRef<Path>>(path: P) -> bool {
    access(path.as_ref(), AccessFlags::F_OK).is_ok()
}

/// Checks if a given path ends with a dot component.
///
/// This function iterates through the bytes of the path from end to start,
/// and determines whether the last component before any slashes is a dot.
///
/// # Arguments
///
/// * `path`: A reference to a `Box<CStr>` representing the path.
///
/// # Returns
///
/// * `bool`: Returns `true` if the path ends with a dot component, otherwise `false`.
///
/// # Examples
///
/// ```ignore
/// use std::ffi::CString;
/// assert_eq!(path_ends_with_dot(&Box::new(CString::new("some/path/.").unwrap())), true);
/// assert_eq!(path_ends_with_dot(&Box::new(CString::new("some/path/foo").unwrap())), false);
/// assert_eq!(path_ends_with_dot(&Box::new(CString::new("some/path/./").unwrap())), true);
/// assert_eq!(path_ends_with_dot(&Box::new(CString::new("some/path/././././///").unwrap())), true);
/// ```
#[allow(clippy::arithmetic_side_effects)]
#[allow(clippy::if_same_then_else)]
#[inline]
pub(crate) fn path_ends_with_dot(path: &CStr) -> bool {
    let bytes = path.to_bytes();

    // Start from the end of the string and move backwards
    let mut index = bytes.len();
    if index == 0 {
        return false;
    }

    // Skip trailing slashes
    while index > 0 && bytes[index - 1] == b'/' {
        index -= 1;
    }

    // If the path is empty after removing trailing slashes, it does not end with a dot
    if index == 0 {
        return false;
    }

    // Check for '.' or '..'
    if bytes[index - 1] == b'.' {
        if index == 1 || bytes[index - 2] == b'/' {
            return true; // Matches '.' or '*/.'
        } else if index > 1 && bytes[index - 2] == b'.' && (index == 2 || bytes[index - 3] == b'/')
        {
            return true; // Matches '..' or '*/..'
        }
    }

    false
}

// Check if the path is a process-referential symbolic link.
pub(crate) fn path_is_selfref_symlink<P: AsRef<Path>>(path: P) -> bool {
    let bytes = path.as_ref().as_os_str().as_bytes();
    matches!(
        bytes,
        b"/proc/self"
            | b"/proc/thread-self"
            | b"/dev/fd"
            | b"/dev/stdin"
            | b"/dev/stdout"
            | b"/dev/stderr"
    )
}

/// Ensure the file we open is a regular file.
/// Ensure we do not block on a fifo if the file is one.
/// Returns (File, size) on success, Errno on failure.
pub fn safe_open<P: AsRef<Path>>(path: P) -> Result<(File, u64), Errno> {
    // Open the file in read-only and non-blocking mode.
    // This way we ensure we don't block even if the file is a FIFO.
    let flags = OFlag::O_RDONLY | OFlag::O_NONBLOCK | OFlag::O_CLOEXEC;
    let fd = open(path.as_ref(), flags, Mode::empty())?;
    // SAFETY: open returns a valid FD.
    let fd = unsafe { OwnedFd::from_raw_fd(fd) };

    // Check the file type and bail if it's not a regular file.
    let stat = fstat(fd.as_raw_fd())?;
    let mode = stat.st_mode;
    if (mode & SFlag::S_IFMT.bits()) != SFlag::S_IFREG.bits() {
        return Err(Errno::ENOENT);
    }

    // Change the file descriptor to blocking mode
    let cur = fcntl(fd.as_raw_fd(), FcntlArg::F_GETFL)?;
    let new = OFlag::from_bits_truncate(cur) & !OFlag::O_NONBLOCK;
    fcntl(fd.as_raw_fd(), FcntlArg::F_SETFL(new))?;

    // Return the file descriptor as a File object
    let file = File::from(fd);
    // SAFETY: no sign loss due to the if-check.
    #[allow(clippy::cast_sign_loss)]
    let size = if stat.st_size >= 0 {
        stat.st_size as u64
    } else {
        0
    };
    Ok((file, size))
}

// Open a path safely using O_PATH and return an OwnedFd.
fn safe_open_path<P: AsRef<Path>>(path: P, flags: OFlag) -> Result<OwnedFd, Errno> {
    let flags = flags | OFlag::O_PATH | OFlag::O_NOFOLLOW;
    let mut how = OpenHow::new(flags.bits(), 0);
    how.resolve = ResolveFlags::NO_MAGICLINKS | ResolveFlags::NO_SYMLINKS;
    openat2(None, path, &how)
        .map(|fd|
            // SAFETY: openat2 returns a valid FD.
            unsafe { OwnedFd::from_raw_fd(fd) })
        .map_err(err2no)
}

/// Check if the given File is executable.
pub(crate) fn is_executable(file: &File) -> bool {
    // SAFETY: Our nix version does not have AtFlags::AT_EACCES (TODO).
    match PathBuf::from("").with_nix_path(|cstr| unsafe {
        nix::libc::faccessat(
            file.as_raw_fd(),
            cstr.as_ptr(),
            nix::libc::X_OK,
            nix::libc::AT_EACCESS | nix::libc::AT_EMPTY_PATH,
        )
    }) {
        Ok(res) => Errno::result(res).map(drop).is_ok(),
        Err(_) => false,
    }
}

/// Checks if the File is a script with a #! and returns the canonical path of the interpreter.
/// SAFETY: If unsafe_ parameter is true:
///   1. Canonicalizes relative interpreter paths.
///   2. Allow ".." in interpreter paths.
///   Typically, none of the two should be necessary.
/// Returns None if the file does not start with a #!.
pub fn binfmt_script(file: &mut File, pid: Pid, unsafe_: bool) -> Result<Option<PathBuf>, Errno> {
    // Read the first two bytes to check for the marker.
    // Careful, the file may not be shorter than two bytes.
    // Note, read_exact handles EINTR transparently.
    let mut marker = [0; 2];
    match file.read_exact(&mut marker) {
        Ok(_) if marker != [b'#', b'!'] => return Ok(None),
        Err(e) if e.kind() == std::io::ErrorKind::UnexpectedEof && is_executable(file) => {
            return Err(Errno::ENOEXEC)
        }
        Err(e) if e.kind() == std::io::ErrorKind::UnexpectedEof => return Err(Errno::EACCES),
        Err(e) => return Err(err2no(e)),
        _ => {}
    };

    // Reading the first 255 - 2 bytes (we already the read the first 2).
    // Since Linux 5.1 255 is the limit of the shebang line.
    let mut buf = Vec::new();
    file.take(255 - 2).read_to_end(&mut buf).map_err(err2no)?;

    let end_of_line = buf.iter().position(|&x| x == b'\n').unwrap_or(buf.len());
    let start_of_command = buf[..end_of_line]
        .iter()
        .position(|&x| !matches!(x, b' ' | b'\t'))
        .unwrap_or(end_of_line);

    let end_of_command = buf[start_of_command..end_of_line]
        .iter()
        .position(|&x| matches!(x, b' ' | b'\t'))
        .map(|pos| pos.saturating_add(start_of_command))
        .unwrap_or(end_of_line);

    let command_path = &buf[start_of_command..end_of_command];
    let command_path = PathBuf::from(OsString::from_vec(command_path.to_vec()));

    Ok(Some(if command_path.is_relative() {
        if !unsafe_ {
            // SAFETY: Refuse to canonicalize relative path.
            // Use ENOSYS rather than ENOEXEC to distinguish from kernel deny.
            // We do the same for !elf binaries with unsupp_binfmt=false.
            return Err(Errno::ENOSYS);
        } else {
            // Handle relative paths wrt. the given process
            let dir = proc_cwd(pid)?;
            canonicalize(
                pid,
                command_path,
                Some(dir),
                true,     // Resolve symbolic links.
                !unsafe_, // SAFETY: Refuse to canonicalize with dotdot in path.
                MissingHandling::Existing,
            )?
        }
    } else {
        canonicalize(
            pid,
            command_path,
            None,
            true,     // Resolve symbolic links.
            !unsafe_, // SAFETY: Refuse to canonicalize with dotdot in path.
            MissingHandling::Existing,
        )?
    }))
}

/// Parse a FD from a Path.
pub(crate) fn parse_fd<P: AsRef<Path>>(path: P) -> Result<RawFd, Errno> {
    let bytes = path.as_ref().as_os_str().as_bytes();

    // Parsing bytes directly to integer
    let mut fd: RawFd = 0;
    for &b in bytes {
        // Check if the byte is a valid digit
        if !b.is_ascii_digit() {
            return Err(Errno::EBADF);
        }
        fd = fd
            .saturating_mul(10)
            .saturating_add(RawFd::from(b.saturating_sub(b'0')));
    }

    Ok(fd)
}

/// Check if a path is `.' or `..'
#[inline]
pub(crate) fn path_is_dot<P: AsRef<Path>>(path: P) -> bool {
    let path = path.as_ref().as_os_str().as_bytes();
    match path.len() {
        1 if path[0] == b'.' => true,
        2 if path[0] == b'.' && path[1] == b'.' => true,
        _ => false,
    }
}

/// Convert a std::io::Error into a nix::Errno.
#[inline]
pub fn err2no(err: std::io::Error) -> Errno {
    err.raw_os_error()
        .map(Errno::from_i32)
        .unwrap_or(Errno::ENOSYS)
}

/// Create file and write the given content.
pub fn cat<P: AsRef<Path>, T: AsRef<[u8]>>(path: P, content: T) -> std::io::Result<()> {
    let mut file = File::create(path)?;
    file.write_all(content.as_ref())?;
    Ok(())
}

/// Make a file executable.
pub fn chmod_x<P: AsRef<Path>>(path: P) -> std::io::Result<()> {
    // Set permissions to make path executable.
    let metadata = metadata(path.as_ref())?;
    let mut permissions = metadata.permissions();
    permissions.set_mode(0o700); // This sets the file executable for the owner (rwx).
    set_permissions(path.as_ref(), permissions)
}

/// Masks all control characters in a path.
///
/// This function takes a path, converts it to a string using the
/// `Display` trait, and replaces any control characters with the '?'
/// character. This is useful for sanitizing paths that might contain
/// control characters which could affect logging systems or terminal
/// output.
///
/// # Arguments
///
/// * `path` - The path to be sanitized, which can be any type that
/// implements `AsRef<Path>`.
///
/// # Returns
///
/// Returns a `String` where all control characters have been replaced
/// by '?'.
pub fn mask_path<P: AsRef<Path>>(path: P) -> String {
    path.as_ref()
        .to_string_lossy()
        .chars()
        .map(|c| if c.is_control() { '?' } else { c })
        .collect()
}

/// Masks all control characters in a C-style string represented by
/// `CStr`.
///
/// This function converts a C-style string to a Rust `String` using the
/// `to_string_lossy` method, which translates non-UTF-8 sequences into
/// Unicode replacement characters. It then iterates over each
/// character, replacing control characters with '?'. This is useful for
/// safely handling and displaying strings from C interfaces that may
/// contain control characters.
///
/// # Arguments
///
/// * `c_str` - A reference to a C-style string (`CStr`) that may
/// include control characters.
///
/// # Returns
///
/// Returns a `String` where all control characters have been replaced by '?'.
pub fn mask_cstr<S: AsRef<CStr>>(c_str: S) -> String {
    c_str
        .as_ref()
        .to_string_lossy()
        .chars()
        .map(|c| if c.is_control() { '?' } else { c })
        .collect()
}

/// Determines if a path contains any Unicode control characters.
///
/// This function checks a path for the presence of control characters
/// as specified in the Unicode standard (U+0000 to U+001F and U+007F to
/// U+009F). It operates directly on the bytes of the path, ensuring
/// efficient processing without converting to a string. This is
/// particularly useful for validating paths that are intended for
/// display or logging, where control characters could cause formatting
/// issues or security vulnerabilities.
///
/// # Arguments
///
/// * `path` - A reference to a path, which can be any type that
/// implements `AsRef<Path>`.
///
/// # Returns
///
/// Returns `true` if the path contains any control characters,
/// otherwise `false`.
#[inline]
pub fn path_has_control_chars<P: AsRef<Path>>(path: P) -> bool {
    path.as_ref()
        .as_os_str()
        .as_bytes()
        .iter()
        .any(|&b| (b <= 0x1F || (0x7F..=0x9F).contains(&b)))
}

#[derive(Debug, PartialEq)]
enum OwningComponent {
    RootDir,
    CurDir,
    ParentDir,
    Normal(OsString),
}

impl<'a> From<Component<'a>> for OwningComponent {
    fn from(comp: Component<'a>) -> Self {
        match comp {
            Component::RootDir => Self::RootDir,
            Component::CurDir => Self::CurDir,
            Component::ParentDir => Self::ParentDir,
            Component::Normal(s) => Self::Normal(s.to_os_string()),
            Component::Prefix(_) => unreachable!(),
        }
    }
}

/// Detects unsafe paths.
/// These paths are denied access regardless of the state of sandboxing.
#[inline]
pub fn path_unsafe<P: AsRef<Path>>(path: P) -> bool {
    let path = path.as_ref().as_os_str().as_bytes();
    matches!(
        path,
        b"/dev/mem"
            | b"/dev/kmem"
            | b"/dev/port"
            //SAFETY: We return an empty file for compat. (see: #106)
            //| b"/proc/cmdline"
            | b"/proc/config"
            | b"/proc/config.gz"
            | b"/proc/dma"
            | b"/proc/interrupts"
            | b"/proc/iomem"
            | b"/proc/ioports"
            | b"/proc/kallsyms"
            | b"/proc/kcore"
            | b"/proc/key-users"
            | b"/proc/keys"
            | b"/proc/modules"
            | b"/proc/mtrr"
            | b"/proc/partitions"
            | b"/proc/softirqs"
            | b"/proc/swaps"
            | b"/proc/sysrq-trigger"
            | b"/proc/vmallocinfo"
            | b"/proc/version"
            | b"/sys/kernel/notes" // Read: https://lwn.net/Articles/962782/
            | b"/sys/kernel/vmcoreinfo"
    ) || path.starts_with(b"/dev/cpu/")
        || path.starts_with(b"/sys/kernel/boot_params/")
        || path.starts_with(b"/sys/kernel/debug/")
        || path.starts_with(b"/sys/kernel/irq/")
        || path.starts_with(b"/sys/kernel/slab/")
        || path.starts_with(b"/sys/kernel/tracing/")
        || path.starts_with(b"/proc/fs/")
        || path.starts_with(b"/proc/irq/")
        || path.starts_with(b"/proc/tty/")
        || (path.starts_with(b"/proc/")
            && (path.ends_with(b"/auxv")
                || path.ends_with(b"/environ")
                || path.ends_with(b"/maps")
                || path.ends_with(b"/mem")
                || path.ends_with(b"/pagemap")
                || path.ends_with(b"/personality")
                || path.ends_with(b"/stack")
                || path.ends_with(b"/syscall")
                || path.ends_with(b"/wchan")))
}

/// Generates a VecDeque of path components, optimized for performance and correctness.
/// - Assumes handling of the initial RootDir is done by the caller.
/// - Directly walks the path bytes to construct components, efficiently skipping multiple leading slashes.
/// - Retains a trailing RootDir if the path ends with a slash, except for the root path itself.
/// - Utilizes SmallVec for efficient allocation and converts to VecDeque for return.
fn path_components<P: AsRef<Path>>(path: P) -> VecDeque<OwningComponent> {
    let path_bytes = path.as_ref().as_os_str().as_bytes();
    let mut components = SmallVec::<[OwningComponent; 32]>::new();

    // Correctly handle multiple leading slashes.
    let mut start = path_bytes
        .iter()
        .position(|&b| b != b'/')
        .unwrap_or(path_bytes.len());

    // Variable to track if we are at the last component
    let mut at_last_component = false;

    while start < path_bytes.len() && !at_last_component {
        let next_slash = path_bytes[start..].iter().position(|&b| b == b'/');
        #[allow(clippy::arithmetic_side_effects)]
        let end = match next_slash {
            Some(pos) => start + pos,
            None => {
                at_last_component = true;
                path_bytes.len()
            }
        };
        let component = &path_bytes[start..end];

        match component {
            b"." => components.push(OwningComponent::CurDir),
            b".." => components.push(OwningComponent::ParentDir),
            _ => components.push(OwningComponent::Normal(OsString::from_vec(
                component.to_vec(),
            ))),
        }

        // Move to the next component, skipping subsequent slashes.
        #[allow(clippy::arithmetic_side_effects)]
        {
            start = end + 1;
        }
    }

    // Check if the original path ends with a slash and is not just a root path "/".
    if path_bytes.ends_with(&[b'/']) && path_bytes.len() > 1 && !at_last_component {
        components.push(OwningComponent::RootDir);
    }

    components.into_iter().collect()
}

/// Return the canonical, absolute form of a path
///
/// This function is a generalization of [`std::fs::canonicalize`] that
/// allows controlling how symbolic links are resolved and how to deal
/// with missing components. It returns the canonical, absolute form of
/// a path.
///
/// The `resolve` is a boolean parameter which controls whether
/// the last component should be resolved or not. Remaining components
/// are always resolved.
///
/// The `nodotdot` is a boolean paramtere which controls whether
/// paths with .. component in them should be denied.
///
/// The `miss_mode` parameter controls how missing components are handled.
#[allow(clippy::cognitive_complexity)]
pub fn canonicalize<P: AsRef<Path>>(
    pid: Pid,
    path: P,
    cwd: Option<P>,
    resolve: bool,
    nodotdot: bool,
    miss_mode: MissingHandling,
) -> Result<PathBuf, Errno> {
    let path = path.as_ref();

    let mut result = if path.is_relative() {
        if let Some(cwd) = cwd {
            cwd.as_ref().to_path_buf()
        } else if path.is_empty() {
            return Err(Errno::ENOENT);
        } else {
            return Err(Errno::EINVAL);
        }
    } else {
        PathBuf::from("/")
    };

    // Used to skip FD parsing for proc_fd.
    #[allow(clippy::arithmetic_side_effects)]
    let pneg = Pid::from_raw(-pid.as_raw());

    // Symbolic Link Loop Detection.
    const SYMLINKS_TO_LOOK_FOR_LOOPS: i32 = 20;
    let mut followed_symlinks = 0;
    let mut visited_files: Option<HashSet<FileInformation>> = None;

    let mut has_to_be_directory = false;
    let mut parts = path_components(path);

    while let Some(part) = parts.pop_front() {
        match part {
            OwningComponent::Normal(p) => {
                result.push(p);
            }
            OwningComponent::RootDir if parts.is_empty() => {
                // Trailing slash
                has_to_be_directory = true;
                continue;
            }
            OwningComponent::RootDir => {
                // RootDir as part of resolving a symlink to an absolute path!
                result.clear();
                result.push("/");
                continue;
            }
            OwningComponent::ParentDir => {
                if nodotdot {
                    return Err(Errno::EACCES);
                }
                result.pop();
                if result.as_os_str().as_bytes() == b"/" {
                    // RootDir as part of walking up path of an earlier symlink.
                    continue;
                }
            }
            OwningComponent::CurDir => {
                // Skip resolving symlink for current directory components.
                continue;
            }
        }

        // Continue if we're not resolving symlinks.
        if !resolve && parts.is_empty() {
            continue;
        }

        match resolve_symlink(pid, &result) {
            Ok(link_path) => {
                // SAFETY: NO_MAGICLINKS
                if proc_fd(pneg, &result)?.is_some()
                    && (link_path.is_relative()
                        || link_path.as_os_str().as_bytes().ends_with(b" (deleted)"))
                {
                    // (a) pipe:42 socket:42 etc, special paths.
                    // (b) file descriptors that point to deleted paths.
                    continue;
                }

                //// Symbolic Link Loop Detection.
                // SAFETY: SYMLINKS_TO_LOOK_FOR_LOOPS is much smaller than u8::MAX.
                #[allow(clippy::arithmetic_side_effects)]
                if followed_symlinks < SYMLINKS_TO_LOOK_FOR_LOOPS {
                    followed_symlinks += 1;
                } else {
                    let stat = FileInformation::from_link(&result)?;
                    if let Some(ref mut visited_files) = visited_files {
                        if !visited_files.insert(stat) {
                            return Err(Errno::ELOOP);
                        }
                    } else {
                        // Allocate the HashSet only when it's really necessary.
                        visited_files = Some(HashSet::from([stat]));
                    }
                }
                ////

                for link_part in link_path.components().rev() {
                    parts.push_front(link_part.into());
                }
                result.pop();
            }
            Err(Errno::EINVAL) => {
                // File is not a symbolic link, continue.
            }
            Err(error) => match miss_mode {
                MissingHandling::Existing => return Err(error),
                MissingHandling::Normal => {
                    if !(parts.is_empty()
                        || (parts.len() == 1 && parts[0] == OwningComponent::RootDir))
                    {
                        return Err(error);
                    }
                }
                _ => {}
            },
        }
    }

    // Deny access to critical paths.
    if path_unsafe(&result) {
        return Err(Errno::EACCES);
    }

    // Raise ENOTDIR if required.
    match miss_mode {
        MissingHandling::Existing => {
            // Last component must exist, return error if it does not.
            safe_open_path(
                &result,
                if has_to_be_directory {
                    OFlag::O_DIRECTORY
                } else {
                    OFlag::empty()
                },
            )?;
        }
        MissingHandling::Normal if has_to_be_directory => {
            if let Err(Errno::ENOTDIR) = safe_open_path(&result, OFlag::O_DIRECTORY) {
                return Err(Errno::ENOTDIR);
            }
        }
        // SAFETY: No check for missing because EEXIST return
        // value may be misused to identify hidden paths.
        _ => {}
    }

    // Preserve trailing slash as necessary
    if has_to_be_directory {
        result.push("");
    }

    Ok(result)
}

#[cfg(test)]
mod tests {
    use std::{
        ffi::CString,
        fs::{self, OpenOptions},
        os::{
            fd::AsRawFd,
            unix::fs::{symlink, OpenOptionsExt},
        },
        process::Command,
        thread::sleep,
        time::{Duration, SystemTime},
    };

    use anyhow::Context;
    use nix::unistd::getpid;

    use super::{MissingHandling::*, *};

    type TestResult = Result<(), Box<dyn std::error::Error>>;

    // A helper function to get the current atime of a file
    fn get_atime<P: AsRef<Path>>(path: P) -> SystemTime {
        let metadata = fs::metadata(path).expect("Failed to get metadata");
        metadata.accessed().expect("Failed to get accessed time")
    }

    // Helper function to assert that the atime of a file or directory has not changed
    fn assert_atime_unchanged<P: AsRef<Path>, F>(path: P, func: F)
    where
        F: FnOnce() -> Result<PathBuf, Errno>,
    {
        let original_atime_f = get_atime(&path);
        let original_atime_p = get_atime(path.as_ref().parent().unwrap());
        sleep(Duration::from_secs(7));
        assert!(
            func().is_ok(),
            "canonicalize {} failed",
            path.as_ref().display()
        );
        // We allow a 1-second tolerance since some filesystems do not have nanosecond precision.
        let new_atime_f = get_atime(&path);
        let new_atime_p = get_atime(path.as_ref().parent().unwrap());
        assert!(
            new_atime_f <= original_atime_f + Duration::new(1, 0),
            "The atime of the file should not have significantly changed."
        );
        assert!(
            new_atime_p <= original_atime_p + Duration::new(1, 0),
            "The atime of the parent dir should not have significantly changed."
        );
    }

    // std::fs::remove_dir_all stops on the first error.
    // we need something more forceful.
    fn remove_dir_all<P: AsRef<Path>>(path: P) -> std::io::Result<()> {
        let status = Command::new("rm")
            .arg("-rf")
            .arg(path.as_ref().to_string_lossy().to_string())
            .status()?;
        if status.success() {
            Ok(())
        } else {
            Err(std::io::Error::new(
                std::io::ErrorKind::Other,
                format!("Failed to remove directory: {}", path.as_ref().display()),
            ))
        }
    }

    // Helper function to create a symlink chain that eventually loops back to an earlier link
    fn setup_symlink_loop(tmp_dir: &tempfile::TempDir, links: &[(&str, &str)]) {
        for &(src, dst) in links {
            let src_path = tmp_dir.path().join(src);

            // Check and create parent directory for the source if necessary
            if let Some(parent) = src_path.parent() {
                if !parent.exists() {
                    fs::create_dir_all(parent).unwrap();
                }
            }

            // The destination is given relative to the source
            let dst_path = Path::new(dst);

            // Avoid creating a symlink if the source already exists
            if src_path.exists() {
                fs::remove_file(&src_path).unwrap();
            }

            // If the destination is an absolute path or starts with '/', we strip the '/' and prefix with tmp_dir
            let full_dst_path = if dst_path.is_absolute() {
                tmp_dir.path().join(dst_path.strip_prefix("/").unwrap())
            } else {
                src_path.parent().unwrap().join(dst_path)
            };

            // Create parent directories for the destination if they don't exist
            if let Some(parent) = full_dst_path.parent() {
                if !parent.exists() {
                    fs::create_dir_all(parent).unwrap();
                }
            }

            // Create the symlink
            symlink(&full_dst_path, &src_path).expect(&format!(
                "Unable to symlink {} -> {}",
                src_path.display(),
                full_dst_path.display()
            ));
        }
    }

    struct EndsWithDotTestCase<'a> {
        path: &'a str,
        test: bool,
    }

    const ENDS_WITH_DOT_TESTS: [EndsWithDotTestCase; 10] = [
        EndsWithDotTestCase {
            path: ".",
            test: true,
        },
        EndsWithDotTestCase {
            path: "..",
            test: true,
        },
        EndsWithDotTestCase {
            path: "...",
            test: false,
        },
        EndsWithDotTestCase {
            path: "/.",
            test: true,
        },
        EndsWithDotTestCase {
            path: "/..",
            test: true,
        },
        EndsWithDotTestCase {
            path: "/...",
            test: false,
        },
        EndsWithDotTestCase {
            path: "foo.",
            test: false,
        },
        EndsWithDotTestCase {
            path: "foo./.",
            test: true,
        },
        EndsWithDotTestCase {
            path: "foo/./././/./",
            test: true,
        },
        EndsWithDotTestCase {
            path: "conftest.dir/././././////",
            test: true,
        },
    ];

    fn tempdir() -> Result<PathBuf, Box<dyn std::error::Error>> {
        let path = Path::new(".syd-test");
        std::fs::create_dir_all(path)?;
        let _ = OpenOptions::new()
            .write(true)
            .create(true)
            .mode(0o600)
            .open(path.join("test"))?;
        Ok(path.to_path_buf())
    }

    #[test]
    fn test_ends_with_dot() {
        for (idx, test) in ENDS_WITH_DOT_TESTS.iter().enumerate() {
            let path = CString::new(test.path).unwrap();
            let ends = path_ends_with_dot(&path);
            assert_eq!(
                test.test, ends,
                "EndsWithDotTestCase {} -> \"{}\": {} != {}",
                idx, test.path, test.test, ends
            );
        }
    }

    #[test]
    fn test_binfmt_script() -> TestResult {
        let pid = Pid::this();
        let rsh = Path::new("/bin/sh")
            .canonicalize()
            .expect("/bin/sh is missing? run for the woods!");
        let ssh = Some(rsh);

        cat("file1", "#!/bin/sh").context("writing binfmt_script to file1 failed!")?;
        chmod_x("file1").context("Failed to set file1 as executable!")?;
        assert_eq!(
            binfmt_script(
                &mut File::open("file1").context("Failed to open file1")?,
                pid,
                true
            )
            .context("binfmt_script on file1 failed!")?,
            ssh
        );

        cat("file1", "#!  /bin/sh").context("writing binfmt_script to file1 failed!")?;
        chmod_x("file1").context("Failed to set file1 as executable!")?;
        assert_eq!(
            binfmt_script(
                &mut File::open("file1").context("Failed to open file1")?,
                pid,
                true
            )
            .context("binfmt_script on file1 failed!")?,
            ssh
        );

        cat("file1", "#!  \t /bin/sh").context("writing binfmt_script to file1 failed!")?;
        chmod_x("file1").context("Failed to set file1 as executable!")?;
        assert_eq!(
            binfmt_script(
                &mut File::open("file1").context("Failed to open file1")?,
                pid,
                true
            )
            .context("binfmt_script on file1 failed!")?,
            ssh
        );

        for invalid_char in ['\x0C' /*\f*/, '\x0B' /*\v*/, '\r'] {
            cat("file1", &format!("#!  \t{invalid_char}/bin/sh"))
                .context("writing binfmt_script to file1 failed!")?;
            chmod_x("file1").context("Failed to set file1 as executable!")?;
            let result = binfmt_script(
                &mut File::open("file1").context("Failed to open file1")?,
                pid,
                true,
            );
            assert!(matches!(result, Err(Errno::ENOENT)), "result:{result:?}");
        }

        Ok(())
    }

    #[test]
    fn test_canonicalize_empty_path() -> TestResult {
        assert_eq!(
            canonicalize(Pid::this(), "", None, true, false, Normal),
            Err(Errno::ENOENT)
        );
        assert_eq!(
            canonicalize(Pid::this(), "", None, true, false, Existing),
            Err(Errno::ENOENT)
        );
        assert_eq!(
            canonicalize(Pid::this(), "", None, true, false, Missing),
            Err(Errno::ENOENT)
        );
        assert_eq!(
            canonicalize(Pid::this(), "", None, false, false, Normal),
            Err(Errno::ENOENT)
        );
        assert_eq!(
            canonicalize(Pid::this(), "", None, false, false, Existing),
            Err(Errno::ENOENT)
        );
        assert_eq!(
            canonicalize(Pid::this(), "", None, false, false, Missing),
            Err(Errno::ENOENT)
        );

        Ok(())
    }

    #[test]
    fn test_canonicalize_repetitive_slashes() -> TestResult {
        let result_test = canonicalize(Pid::this(), "/etc/passwd", None, true, false, Normal)?;
        let paths = vec![
            "/etc/passwd",
            "/etc//passwd",
            "/etc///passwd",
            "//etc/passwd",
            "//etc//passwd",
            "//etc///passwd",
            "///etc/passwd",
            "///etc//passwd",
            "///etc///passwd",
        ];
        for path in &paths {
            let result = canonicalize(Pid::this(), path, None, true, false, Normal)?;
            assert_eq!(result, result_test);
        }

        Ok(())
    }

    #[test]
    fn test_canonicalize_dots_slashes() -> TestResult {
        let base = tempdir()?;

        let cwd = std::env::current_dir()?.display().to_string();
        let path = base.display().to_string();

        let result1 = canonicalize(
            Pid::this(),
            format!("{path}//./..//{path}/test"),
            Some(cwd.clone()),
            true,
            false,
            Normal,
        )?
        .display()
        .to_string();
        let result2 = canonicalize(
            Pid::this(),
            format!("{cwd}/{path}//./..//{path}/test"),
            None,
            true,
            false,
            Existing,
        )?
        .display()
        .to_string();

        assert!(!result1.is_empty(), "result:{result1}");
        assert!(!result2.is_empty(), "result:{result2}");
        assert_eq!(result1, result2);

        Ok(())
    }

    #[test]
    fn test_canonicalize_non_directory_with_slash() -> TestResult {
        let cwd = std::env::current_dir()?.display().to_string();
        let path = tempdir()?.display().to_string();
        let test = format!("{path}/test/");

        assert_eq!(
            canonicalize(Pid::this(), &test, Some(&cwd), true, false, Normal),
            Err(Errno::ENOTDIR)
        );
        assert_eq!(
            canonicalize(Pid::this(), &test, Some(&cwd), true, false, Existing),
            Err(Errno::ENOTDIR)
        );
        //SAFETY: Missing is handled by read_path*.
        assert!(canonicalize(Pid::this(), &test, Some(&cwd), true, false, Missing).is_ok());
        assert_eq!(
            canonicalize(Pid::this(), &test, Some(&cwd), false, false, Normal),
            Err(Errno::ENOTDIR)
        );
        assert_eq!(
            canonicalize(Pid::this(), &test, Some(&cwd), false, false, Existing),
            Err(Errno::ENOTDIR)
        );
        //SAFETY: Missing is handled by read_path*.
        assert!(canonicalize(Pid::this(), &test, Some(&cwd), false, false, Missing).is_ok());

        Ok(())
    }

    /// FIXME: The asserts return success rather than failure.
    /// Bug or feature?
    #[test]
    #[ignore]
    fn test_canonicalize_missing_directory_returns_enoent() -> TestResult {
        assert_eq!(
            canonicalize(Pid::this(), "/zzz/..", None, true, false, Normal),
            Err(Errno::ENOENT)
        );
        assert_eq!(
            canonicalize(Pid::this(), "/zzz/..", None, true, false, Existing),
            Err(Errno::ENOENT)
        );
        assert_eq!(
            canonicalize(Pid::this(), "/zzz/..", None, false, false, Normal),
            Err(Errno::ENOENT)
        );
        assert_eq!(
            canonicalize(Pid::this(), "/zzz/..", None, false, false, Existing),
            Err(Errno::ENOENT)
        );

        Ok(())
    }

    #[test]
    fn test_relative_symlink_resolution() -> TestResult {
        // Setup
        let root_test_dir = Path::new("test_root_relative_symlink_resolution");
        let deep_dir = root_test_dir.join("a/b/c");
        let _ = remove_dir_all(&root_test_dir);
        fs::create_dir_all(&root_test_dir.join("d"))?;
        fs::create_dir_all(&deep_dir)?;

        // Create a symlink in "b" that points upwards to "a"
        let rel_link = root_test_dir.join("a/b/rel_link");
        symlink("../..", &rel_link)?;

        // Append /proc/self/cwd to get an absolute path to our symlinked path
        let abs_link_path = Path::new("/proc/self/cwd").join(root_test_dir.join("a/b/rel_link/d"));

        // Call canonicalize
        let result = canonicalize(
            Pid::this(),
            &abs_link_path,
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert!(
            matches!(result, Ok(_)),
            "canonicalize:{} result:{:?}",
            abs_link_path.display(),
            result
        );
        let resolved_path = result.unwrap();

        // We expect the path to be resolved to "test_root/a/d", but we need to canonicalize it
        let expected_path =
            fs::canonicalize(Path::new("/proc/self/cwd").join(root_test_dir.join("d")))?;

        // Cleanup
        let _ = remove_dir_all(&root_test_dir);

        assert_eq!(resolved_path, expected_path);

        Ok(())
    }

    // FIXME: This test broke after we removed normalize()
    // The question: Is the test incorrect or is canonicalize()?
    #[ignore]
    #[test]
    fn test_complex_interplay_symlinks_dots() -> TestResult {
        // Setup
        let cwd = Path::new("/proc/self/cwd").canonicalize()?;
        let root_test_dir = cwd.join("test_root_complex_interplay_symlinks_dots");
        let _ = remove_dir_all(&root_test_dir);
        fs::create_dir_all(root_test_dir.join("a/b/c")).unwrap();
        fs::create_dir(root_test_dir.join("d")).unwrap();
        fs::create_dir(root_test_dir.join("e")).unwrap();
        fs::create_dir(root_test_dir.join("x")).unwrap();

        // Create several symlinks
        symlink("./a", root_test_dir.join("link_to_a")).unwrap();
        symlink("e", root_test_dir.join("link_to_e")).unwrap();
        symlink("a/b", root_test_dir.join("link_to_b")).unwrap();
        symlink("../../x", root_test_dir.join("a/b/rel_link")).unwrap();

        let path = root_test_dir.join("link_to_a/../link_to_b/rel_link/../..");
        let resolved_path = canonicalize(
            Pid::this(),
            path,
            None,
            true,
            false,
            MissingHandling::Existing,
        )
        .unwrap();

        // Cleanup
        let _ = remove_dir_all(&root_test_dir);

        // Assertion
        assert_eq!(resolved_path, root_test_dir);

        Ok(())
    }

    #[test]
    fn test_trailing_slash_handling() -> TestResult {
        let path = Path::new("/usr/");
        let pabs = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Normal,
        )
        .unwrap();
        assert_eq!(path, pabs);

        let path = Path::new("/proc/self/");
        let pexp = format!("/proc/{}/", getpid());
        let pexp = Path::new(&pexp);
        let pabs = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Normal,
        )
        .unwrap();
        assert_eq!(pabs, pexp);
        let pabs = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Normal,
        )
        .unwrap();
        assert_eq!(pabs, pexp);

        Ok(())
    }

    #[ignore]
    #[test]
    fn test_canonicalize_no_atime_change_normal() -> TestResult {
        let cdir = std::env::current_dir()?;
        let base = cdir.join(tempdir()?);
        let path = base.join("file");
        fs::File::create(&path)?;

        assert_atime_unchanged(&path, || {
            canonicalize(
                Pid::this(),
                &path,
                None,
                true,
                false,
                MissingHandling::Normal,
            )
        });

        let _ = remove_dir_all(&base);
        Ok(())
    }

    #[ignore]
    #[test]
    fn test_canonicalize_no_atime_change_existing() -> TestResult {
        let cdir = std::env::current_dir()?;
        let base = cdir.join(tempdir()?);
        let path = base.join("file");
        fs::File::create(&path)?;

        assert_atime_unchanged(&path, || {
            canonicalize(
                Pid::this(),
                &path,
                None,
                true,
                false,
                MissingHandling::Existing,
            )
        });

        let _ = remove_dir_all(&base);
        Ok(())
    }

    #[test]
    fn test_canonicalize_symlink_loop() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let dir_path = tmp_dir.path();

        // Create a symlink loop: link_a -> link_b -> link_a
        let mut link_a = dir_path.join("link_a");
        let mut link_b = dir_path.join("link_b");
        symlink(&link_b, &link_a).expect("Failed to create symlink a");
        symlink(&link_a, &link_b).expect("Failed to create symlink b");

        // Now check that canonicalize detects the loop correctly
        let result = canonicalize(
            Pid::this(),
            &link_a,
            None,
            false,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &link_a,
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &link_a,
            None,
            false,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &link_a,
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &link_a,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &link_a,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a trailing slash and retest.
        link_a.push("");

        let result = canonicalize(
            Pid::this(),
            &link_a,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &link_a,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a trailing slash and retest.
        link_b.push("");

        let result = canonicalize(
            Pid::this(),
            &link_b,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &link_b,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));
    }

    #[test]
    fn test_canonicalize_nonexistent_final_component() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let dir_path = tmp_dir.path();

        // Create a valid symlink to a non-existent final component
        let mut valid_link = dir_path.join("valid_link");
        let nonexistent_target = dir_path.join("nonexistent");
        symlink(&nonexistent_target, &valid_link)
            .expect("Failed to create symlink to non-existent target");

        // Now check that canonicalize handles the non-existent final component correctly
        let result = canonicalize(
            Pid::this(),
            &valid_link,
            None,
            false,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &valid_link,
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &valid_link,
            None,
            false,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &valid_link,
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert_eq!(result, Err(Errno::ENOENT));

        let result = canonicalize(
            Pid::this(),
            &valid_link,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");

        // FIXME: Why oh why?
        let result = canonicalize(
            Pid::this(),
            &valid_link,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert!(result.is_ok(), "{result:?}");
        // should be: assert_eq!(result, Err(Errno::EEXIST));

        // Add a trailing slash and retest.
        valid_link.push("");

        // FIXME: Why oh why?
        let result = canonicalize(
            Pid::this(),
            &valid_link,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        assert!(result.is_ok(), "{result:?}");
        // should be: assert_eq!(result, Err(Errno::EEXIST));

        // FIXME: Why oh why?
        let result = canonicalize(
            Pid::this(),
            &valid_link,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert!(result.is_ok(), "{result:?}");
        // should be: assert_eq!(result, Err(Errno::EEXIST));
    }

    #[test]
    fn test_canonicalize_self_referential_symlink() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let mut symlink_path = tmp_dir.path().join("self_link");
        symlink(&symlink_path, &symlink_path).expect("Failed to create self-referential symlink");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            false,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            false,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a trailing slash and retest.
        symlink_path.push("");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));
    }

    #[test]
    fn test_canonicalize_broken_symlink() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let mut broken_link = tmp_dir.path().join("broken_link");
        let nonexistent_target = tmp_dir.path().join("nonexistent_target");
        symlink(&nonexistent_target, &broken_link).expect("Failed to create broken symlink");

        let result = canonicalize(
            Pid::this(),
            &broken_link,
            None,
            false,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &broken_link,
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        // FIXME: assert_eq!(result, Err(Errno::ENOENT));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &broken_link,
            None,
            false,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &broken_link,
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert_eq!(result, Err(Errno::ENOENT));

        let result = canonicalize(
            Pid::this(),
            &broken_link,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &broken_link,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert!(result.is_ok(), "{result:?}");

        // Add a trailing slash and retest.
        broken_link.push("");

        let result = canonicalize(
            Pid::this(),
            &broken_link,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &broken_link,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert!(result.is_ok(), "{result:?}");
    }

    #[test]
    fn test_canonicalize_symlink_to_directory() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let dir = tmp_dir.path().join("dir");
        fs::create_dir(&dir).expect("Failed to create directory");

        let symlink_path = tmp_dir.path().join("dir_link");
        symlink(&dir, &symlink_path).expect("Failed to create symlink to directory");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            false,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_dir(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            false,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_dir(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &symlink_path,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");
    }

    #[test]
    fn test_canonicalize_symlink_chain() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");
        let link1 = tmp_dir.path().join("link1");
        let link2 = tmp_dir.path().join("link2");
        let link3 = tmp_dir.path().join("link3");
        let file = tmp_dir.path().join("file");
        fs::write(&file, "content").expect("Failed to write file");

        // Create a chain of symlinks: link1 -> link2 -> link3 -> file
        symlink(&link2, &link1).expect("Failed to create link1");
        symlink(&link3, &link2).expect("Failed to create link2");
        symlink(&file, &link3).expect("Failed to create link3");

        let result = canonicalize(
            Pid::this(),
            &link1,
            None,
            false,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &link1,
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_file(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &link1,
            None,
            false,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &link1,
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_file(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &link1,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &link1,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");
    }

    #[test]
    fn test_canonicalize_complex_symlink_loop_with_intermediate_components() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");

        // Setting up a complex symlink scenario
        setup_symlink_loop(
            &tmp_dir,
            &[("a", "b/c"), ("b/c", "d"), ("b/d", "../e"), ("e", "f/../a")],
        );

        let mut path = tmp_dir.path().join("a");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(result.clone().unwrap().is_symlink(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a trailing slash and retest.
        path.push("");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a final component and retest.
        path.push("foo");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));
    }

    #[test]
    fn test_canonicalize_symlinks_with_dot_and_dotdot_components() {
        let tmp_dir = tempfile::tempdir().expect("Failed to create temp dir");

        // Create a complex directory structure with dots and symlinks
        fs::create_dir_all(tmp_dir.path().join("b")).expect("Failed to create directory b");
        symlink("b", tmp_dir.path().join("a")).expect("Failed to create symlink a -> b");
        symlink("..///e", tmp_dir.path().join("b").join("d"))
            .expect("Failed to create symlink b/d -> ../e");
        symlink("b/.///./d", tmp_dir.path().join("e")).expect("Failed to create symlink e -> b/d");

        let mut path = tmp_dir.path().join("a").join(".").join("d");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(!result.clone().unwrap().exists(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");
        assert!(!result.clone().unwrap().exists(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a trailing slash and retest.
        path.push("");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        // Add a final component and retest.
        path.push("foo");

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            false,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));

        let result = canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        assert_eq!(result, Err(Errno::ELOOP));
    }

    #[test]
    fn test_canonicalize_proc_self() -> TestResult {
        let path = canonicalize(
            Pid::this(),
            "/proc/self",
            None,
            false,
            false,
            MissingHandling::Normal,
        )?;
        assert_eq!(path, PathBuf::from("/proc/self"));

        Ok(())
    }

    #[test]
    fn test_canon_glob_std() -> TestResult {
        let result = canonicalize(
            Pid::this(),
            "/proc/self/fd/0",
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");
        let result = canonicalize(
            Pid::this(),
            "/proc/self/fd/1",
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");
        let result = canonicalize(
            Pid::this(),
            "/proc/self/fd/2",
            None,
            true,
            false,
            MissingHandling::Normal,
        );
        assert!(result.is_ok(), "{result:?}");
        let result = canonicalize(
            Pid::this(),
            "/proc/self/fd/0",
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");
        let result = canonicalize(
            Pid::this(),
            "/proc/self/fd/1",
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");
        let result = canonicalize(
            Pid::this(),
            "/proc/self/fd/2",
            None,
            true,
            false,
            MissingHandling::Existing,
        );
        assert!(result.is_ok(), "{result:?}");
        let result = canonicalize(
            Pid::this(),
            "/proc/self/fd/0",
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");
        let result = canonicalize(
            Pid::this(),
            "/proc/self/fd/1",
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");
        let result = canonicalize(
            Pid::this(),
            "/proc/self/fd/2",
            None,
            true,
            false,
            MissingHandling::Missing,
        );
        //SAFETY: Missing is handled by read_path*.
        //assert_eq!(result, Err(Errno::EEXIST));
        assert!(result.is_ok(), "{result:?}");
        Ok(())
    }

    #[test]
    fn test_canon_glob_pipe() -> TestResult {
        let (read_end, write_end) = std::os::unix::net::UnixStream::pair()?;

        let fd = read_end.as_raw_fd();
        let path = format!("/proc/self/fd/{fd}");
        assert!(canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Normal
        )
        .is_ok());

        let fd = write_end.as_raw_fd();
        let path = format!("/proc/self/fd/{fd}");
        assert!(canonicalize(
            Pid::this(),
            &path,
            None,
            true,
            false,
            MissingHandling::Normal
        )
        .is_ok());

        Ok(())
    }
}
