//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/proc.rs: /proc utilities
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
// Based in part upon procps-ng's library/devname.c which is:
//   Copyright © 2011-2023 Jim Warner <james.warner@comcast.net>
//   Copyright © 2017-2023 Craig Small <csmall@dropbear.xyz>
//   Copyright © 1998-2009 Albert Cahalan
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Used only by proc_cmdline.
#[cfg(feature = "log")]
use std::ffi::OsString;
#[cfg(feature = "log")]
use std::os::unix::ffi::OsStringExt;
use std::{
    env::VarError,
    ffi::OsStr,
    fs::{read_dir, read_link, File},
    io::{Read, Write},
    os::{fd::RawFd, unix::ffi::OsStrExt},
    path::{Path, PathBuf},
};

use nix::{errno::Errno, libc::pid_t, sys::stat::Mode, unistd::Pid};
use procfs::process::{MMapPath, Process};
use smallvec::SmallVec;

const TGID_PREFIX: &[u8] = b"Tgid:\t";
const UMASK_PREFIX: &[u8] = b"Umask:\t";

// major 204 is a mess -- "Low-density serial ports"
const LOW_DENSITY_NAMES: &[&str] = &[
    "LU0", "LU1", "LU2", "LU3", "FB0", "SA0", "SA1", "SA2", "SC0", "SC1", "SC2", "SC3", "FW0",
    "FW1", "FW2", "FW3", "AM0", "AM1", "AM2", "AM3", "AM4", "AM5", "AM6", "AM7", "AM8", "AM9",
    "AM10", "AM11", "AM12", "AM13", "AM14", "AM15", "DB0", "DB1", "DB2", "DB3", "DB4", "DB5",
    "DB6", "DB7", "SG0", "SMX0", "SMX1", "SMX2", "MM0", "MM1", "CPM0", "CPM1", "CPM2",
    "CPM3", /* "CPM4", "CPM5", */
    // bad allocation?
    "IOC0", "IOC1", "IOC2", "IOC3", "IOC4", "IOC5", "IOC6", "IOC7", "IOC8", "IOC9", "IOC10",
    "IOC11", "IOC12", "IOC13", "IOC14", "IOC15", "IOC16", "IOC17", "IOC18", "IOC19", "IOC20",
    "IOC21", "IOC22", "IOC23", "IOC24", "IOC25", "IOC26", "IOC27", "IOC28", "IOC29", "IOC30",
    "IOC31", "VR0", "VR1", "IOC84", "IOC85", "IOC86", "IOC87", "IOC88", "IOC89", "IOC90", "IOC91",
    "IOC92", "IOC93", "IOC94", "IOC95", "IOC96", "IOC97", "IOC98", "IOC99", "IOC100", "IOC101",
    "IOC102", "IOC103", "IOC104", "IOC105", "IOC106", "IOC107", "IOC108", "IOC109", "IOC110",
    "IOC111", "IOC112", "IOC113", "IOC114", "IOC115", "SIOC0", "SIOC1", "SIOC2", "SIOC3", "SIOC4",
    "SIOC5", "SIOC6", "SIOC7", "SIOC8", "SIOC9", "SIOC10", "SIOC11", "SIOC12", "SIOC13", "SIOC14",
    "SIOC15", "SIOC16", "SIOC17", "SIOC18", "SIOC19", "SIOC20", "SIOC21", "SIOC22", "SIOC23",
    "SIOC24", "SIOC25", "SIOC26", "SIOC27", "SIOC28", "SIOC29", "SIOC30", "SIOC31", "PSC0", "PSC1",
    "PSC2", "PSC3", "PSC4", "PSC5", "AT0", "AT1", "AT2", "AT3", "AT4", "AT5", "AT6", "AT7", "AT8",
    "AT9", "AT10", "AT11", "AT12", "AT13", "AT14", "AT15", "NX0", "NX1", "NX2", "NX3", "NX4",
    "NX5", "NX6", "NX7", "NX8", "NX9", "NX10", "NX11", "NX12", "NX13", "NX14", "NX15",
    "J0", // minor is 186
    "UL0", "UL1", "UL2", "UL3", "xvc0", // FAIL -- "/dev/xvc0" lacks "tty" prefix
    "PZ0", "PZ1", "PZ2", "PZ3", "TX0", "TX1", "TX2", "TX3", "TX4", "TX5", "TX6", "TX7", "SC0",
    "SC1", "SC2", "SC3", "MAX0", "MAX1", "MAX2", "MAX3",
];

/// Retrieves the thread group ID (TGID) for the specified thread ID (TID).
pub fn proc_tgid(tid: Pid) -> Result<Pid, Errno> {
    // Construct path to the appropriate status file
    let mut buf = itoa::Buffer::new();
    let mut proc = PathBuf::from("/proc");
    proc.push(buf.format(tid.as_raw()));
    proc.push("status");

    // Open the file
    let mut file = File::open(proc).map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })?;
    let mut buf = [0; 84]; // Read the first 84 bytes where the tgid is likely to be.
    let bytes_read = file.read(&mut buf).map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })?;

    // Search for "Tgid:" pattern directly in bytes
    if let Some(position) = buf[..bytes_read]
        .windows(TGID_PREFIX.len())
        .position(|window| window == TGID_PREFIX)
    {
        let start = position.saturating_add(TGID_PREFIX.len());
        // Find the end of the tgid value assuming it ends with a newline.
        if let Some(end) = buf[start..].iter().position(|&b| b == b'\n') {
            // Parse tgid directly from bytes
            let mut tgid: pid_t = 0;
            for &digit in &buf[start..start.saturating_add(end)] {
                tgid = tgid
                    .saturating_mul(10)
                    .saturating_add(pid_t::from(digit.saturating_sub(b'0')));
            }
            return Ok(Pid::from_raw(tgid));
        }
    }

    Err(Errno::ESRCH)
}

/// Get the umask of the given `Pid`.
pub(crate) fn proc_umask(pid: Pid) -> Result<Mode, Errno> {
    // Construct path to the appropriate status file
    let mut buf = itoa::Buffer::new();
    let mut proc = PathBuf::from("/proc");
    proc.push(buf.format(pid.as_raw()));
    proc.push("status");

    // Open the file
    let mut file = File::open(proc).map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })?;
    let mut buf = [0u8; 42]; // Read the first 42 bytes where the umask is likely to be.
    let bytes_read = file.read(&mut buf).map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })?;

    // Search for "Umask:" pattern directly in bytes
    if let Some(position) = buf[..bytes_read]
        .windows(UMASK_PREFIX.len())
        .position(|window| window == UMASK_PREFIX)
    {
        let start = position.saturating_add(UMASK_PREFIX.len());
        // Find the end of the umask value assuming it ends with a newline.
        if let Some(end) = buf[start..].iter().position(|&b| b == b'\n') {
            // Parse umask directly from bytes
            let mut umask = 0u32;
            for &digit in &buf[start..start.saturating_add(end)] {
                umask = umask
                    .saturating_mul(8)
                    .saturating_add(u32::from(digit.saturating_sub(b'0')));
            }
            return Ok(Mode::from_bits_truncate(umask));
        }
    }

    Err(Errno::EIO)
}

/// Retrieves the command line of the specified process ID (PID)
/// concatenated as a single string.
///
/// This function reads the `/proc/<pid>/cmdline` file and concatenates
/// the arguments using spaces. The function takes care of replacing null
/// bytes (`'\0'`) with spaces to format the command line as a readable string.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the command line.
///
/// # Returns
///
/// * `Ok(String)` - The command line of the process as a single string.
/// * `Err(Errno)` - An error, depending on the issue encountered while reading
///                  or processing the cmdline file.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_cmdline;
///
/// let cmdline = proc_cmdline(Pid::this());
/// assert!(cmdline.is_ok());
/// ```
#[cfg(feature = "log")]
pub fn proc_cmdline(pid: Pid) -> Result<OsString, Errno> {
    const LIMIT: usize = 256;

    // Construct path to the appropriate cmdline file.
    let mut buf = itoa::Buffer::new();
    let mut cmd = PathBuf::from("/proc");
    cmd.push(buf.format(pid.as_raw()));
    cmd.push("cmdline");

    let mut file = File::open(cmd).map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })?;
    let mut buf = vec![0u8; LIMIT];
    let bytes_read = file.read(&mut buf).map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })?;

    if bytes_read == 0 {
        return Ok(OsString::new()); // empty cmdline
    }

    // Determine if EOF was reached or if we hit the limit
    if bytes_read == LIMIT {
        // Check if the last byte read is not a null byte, indicating there's more data
        if buf[LIMIT - 1] != 0 {
            // Append ellipsis to indicate truncation
            buf.extend_from_slice("…".as_bytes());
        }
    } else {
        // If EOF was hit before the limit, resize the buffer to bytes_read
        buf.resize(bytes_read, 0);
    }

    // Replace null bytes with spaces
    for byte in &mut buf {
        if *byte == 0 {
            *byte = b' ';
        }
    }

    Ok(OsString::from_vec(buf))
}

/// Retrieves the command name (comm) of the specified process ID (PID)
/// as a single string.
///
/// This function reads the `/proc/<pid>/comm` file. It reads up to 16 characters,
/// which is typically sufficient for process names.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the command name.
///
/// # Returns
///
/// * `Ok(String)` - The command name of the process as a single string.
/// * `Err(Errno)` - An error, depending on the issue encountered while reading
///                  the comm file.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_comm;
///
/// let comm = proc_comm(Pid::this());
/// assert!(comm.is_ok());
/// ```
#[cfg(not(feature = "log"))]
pub fn proc_comm(pid: Pid) -> Result<String, Errno> {
    const LIMIT: usize = 16;
    let mut buf = itoa::Buffer::new();
    let mut path = PathBuf::from("/proc");
    path.push(buf.format(pid.as_raw()));
    path.push("comm");

    let mut file = File::open(path).map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })?;
    let mut data = Vec::with_capacity(LIMIT);

    // Read up to LIMIT characters or until EOF
    file.read_to_end(&mut data).map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })?;

    let comm = String::from_utf8_lossy(&data);
    Ok(comm.trim_end().to_string())
}

/// Retrieves the current working directory (CWD) of the specified process ID (PID).
///
/// This function reads the symbolic link `/proc/<pid>/cwd` to determine the CWD.
///
/// # Arguments
///
/// * `pid` - The process ID for which to retrieve the current working directory.
///
/// # Returns
///
/// * `Ok(PathBuf)` - The current working directory of the process.
/// * `Err(Errno)` - An error, depending on the issue encountered while reading
///                  or resolving the `cwd` symlink.
///
/// # Examples
///
/// ```rust
/// use nix::unistd::Pid;
/// use syd::proc::proc_cwd;
///
/// let cwd = proc_cwd(Pid::this());
/// assert!(cwd.is_ok());
/// ```
pub fn proc_cwd(pid: Pid) -> Result<PathBuf, Errno> {
    let mut buf = itoa::Buffer::new();
    let mut path = PathBuf::from("/proc");
    path.push(buf.format(pid.as_raw()));
    path.push("cwd");
    read_link(path).map(PathBuf::from).map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })
}

/// Reads the tty number from /proc/[pid]/stat and figures out the corresponding /dev/tty device node path.
#[allow(clippy::arithmetic_side_effects)]
#[allow(clippy::cast_sign_loss)]
#[allow(clippy::disallowed_methods)]
pub fn proc_tty(pid: Pid) -> Result<PathBuf, Errno> {
    let stat = Process::new(pid.as_raw())
        .map_err(|_| Errno::ENOENT)?
        .stat()
        .map_err(|_| Errno::EPERM)?;
    if stat.tty_nr <= 0 {
        // Process has no controlling terminal
        return Err(Errno::ENXIO);
    }
    let (major, minor) = stat.tty_nr();
    match major {
        3 => Ok(PathBuf::from(format!(
            "/dev/tty{}{}",
            "pqrstuvwxyzabcde"[(minor >> 4) as usize..]
                .chars()
                .next()
                .unwrap(),
            "0123456789abcdef"[(minor & 0x0f) as usize..]
                .chars()
                .next()
                .unwrap()
        ))),
        4 => {
            if minor < 64 {
                Ok(PathBuf::from(format!("/dev/tty{}", minor)))
            } else {
                Ok(PathBuf::from(format!("/dev/ttyS{}", minor - 64)))
            }
        }
        11 => Ok(PathBuf::from(format!("/dev/ttyB{}", minor))),
        14 => Ok(PathBuf::from(format!("/dev/tty{}", minor))), // Standard TTYs
        17 => Ok(PathBuf::from(format!("/dev/ttyH{}", minor))),
        19 | 22 | 23 => Ok(PathBuf::from(format!("/dev/ttyD{}", minor))),
        24 => Ok(PathBuf::from(format!("/dev/ttyE{}", minor))),
        32 => Ok(PathBuf::from(format!("/dev/ttyX{}", minor))),
        43 => Ok(PathBuf::from(format!("/dev/ttyI{}", minor))),
        46 => Ok(PathBuf::from(format!("/dev/ttyR{}", minor))),
        48 => Ok(PathBuf::from(format!("/dev/ttyL{}", minor))),
        57 => Ok(PathBuf::from(format!("/dev/ttyP{}", minor))),
        71 => Ok(PathBuf::from(format!("/dev/ttyF{}", minor))),
        75 => Ok(PathBuf::from(format!("/dev/ttyW{}", minor))),
        78 | 112 => Ok(PathBuf::from(format!("/dev/ttyM{}", minor))),
        105 => Ok(PathBuf::from(format!("/dev/ttyV{}", minor))),
        136..=143 => Ok(PathBuf::from(format!(
            "/dev/pts/{}",
            minor + (major - 136) * 256
        ))),
        148 => Ok(PathBuf::from(format!("/dev/ttyT{}", minor))),
        154 | 156 => Ok(PathBuf::from(format!(
            "/dev/ttySR{}",
            minor + if major == 156 { 256 } else { 0 }
        ))),
        164 => Ok(PathBuf::from(format!("/dev/ttyCH{}", minor))),
        166 => Ok(PathBuf::from(format!("/dev/ttyACM{}", minor))),
        172 => Ok(PathBuf::from(format!("/dev/ttyMX{}", minor))),
        174 => Ok(PathBuf::from(format!("/dev/ttySI{}", minor))),
        188 => Ok(PathBuf::from(format!("/dev/ttyUSB{}", minor))),
        204 => {
            if minor as usize >= LOW_DENSITY_NAMES.len() {
                Err(Errno::ENXIO)
            } else {
                Ok(PathBuf::from(format!(
                    "/dev/tty{}",
                    LOW_DENSITY_NAMES[minor as usize]
                )))
            }
        }
        208 => Ok(PathBuf::from(format!("/dev/ttyU{}", minor))),
        216 => Ok(PathBuf::from(format!("/dev/ttyUB{}", minor))),
        224 => Ok(PathBuf::from(format!("/dev/ttyY{}", minor))),
        227 => Ok(PathBuf::from(format!("/dev/3270/tty{}", minor))),
        229 => Ok(PathBuf::from(format!("/dev/iseries/vtty{}", minor))),
        256 => Ok(PathBuf::from(format!("/dev/ttyEQ{}", minor))),
        _ => Err(Errno::ENXIO),
    }
}

/// Sets the maximum number of user namespaces.
///
/// This function opens the file `/proc/sys/user/max_user_namespaces`
/// and writes the value `1` into it. It is used to limit the number of
/// user namespaces that can be created.
///
/// # Returns
/// * `Ok(())` on success.
/// * `Err(Errno)` on failure, with the error converted from `std::io::Error`.
pub fn proc_limit_userns() -> Result<(), Errno> {
    match File::create("/proc/sys/user/max_user_namespaces") {
        Ok(mut file) => match file.write_all(b"1") {
            Ok(_) => Ok(()),
            Err(e) => Err(e.raw_os_error().map(Errno::from_i32).unwrap_or(Errno::EIO)),
        },
        Err(e) => Err(e.raw_os_error().map(Errno::from_i32).unwrap_or(Errno::EIO)),
    }
}

/// Checks if the number of tasks across all processes in the system
/// exceeds the given limit by inspecting the `/proc` filesystem.
///
/// This function iterates over each entry in the `/proc` directory,
/// assuming each numeric directory name represents a process ID. For
/// each process, it counts the number of tasks (threads) by counting
/// entries in the `/proc/[pid]/task` directory. The function skips
/// over the initial entry `/proc/1`.
///
/// # Returns
/// `Ok(true)`: Number of tasks reach the given limit.
/// `Err(Errno)`: An error occurred during the operation, excluding
/// ENOENT (file not found), EACCES (access denied), EPERM (permission
/// denied), and ESRCH (process not found) errors, which are ignored.
///
/// # Errors
/// This function returns an error for any issue encountered while
/// reading the `/proc` directory or task subdirectories, except for
/// ENOENT, EACCES, EPERM, and ESRCH errors, which are ignored. The
/// error is wrapped in an `Errno` type.
pub fn proc_task_limit(max: usize) -> Result<bool, Errno> {
    let proc_entries = read_dir("/proc").map_err(|err| {
        err.raw_os_error()
            .map(Errno::from_i32)
            .unwrap_or(Errno::ENOENT)
    })?;

    let result = proc_entries
        .filter_map(Result::ok)
        .filter(|entry| entry.file_name().as_bytes() != b"1" && is_numeric(&entry.file_name()))
        .try_fold(
            (0usize, SmallVec::<[PathBuf; 256]>::new()),
            |(count, mut entries), entry| {
                let count = count.saturating_add(1);
                if count >= max {
                    // Early termination with Err to break out of the try_fold
                    return Err(());
                }
                entries.push(entry.path());
                Ok((count, entries))
            },
        );

    // Unwrap the result of try_fold, handling early termination
    let (mut pid_count, eligible_entries) = match result {
        Ok(data) => data,
        Err(_) => return Ok(true),
    };

    // Step 2: Count tasks in each PID
    for path in eligible_entries {
        match read_dir(&path.join("task")) {
            Ok(tasks) => {
                pid_count = pid_count.saturating_add(tasks.count()).saturating_sub(1);
                if pid_count >= max {
                    return Ok(true);
                }
            }
            Err(error) => match error.raw_os_error() {
                Some(
                    nix::libc::ENOENT | nix::libc::EPERM | nix::libc::EACCES | nix::libc::ESRCH,
                ) => {}
                Some(error) => return Err(Errno::from_i32(error)),
                None => return Err(Errno::ENOENT),
            },
        }
    }

    Ok(false)
}

#[inline]
fn is_numeric(name: &OsStr) -> bool {
    name.as_bytes().iter().all(|&b| b.is_ascii_digit())
}

const FD: &[u8] = b"/fd/";
const PROC: &[u8] = b"/proc/";
const TASK: &[u8] = b"/task/";

/// 0. Assumes the given path is normalized.
/// 1. The path must start with /proc.
/// 2. The second component must be a numeric PID equal to the given PID.
/// 3. An optional task/[TID] part, where [TID] is a numeric PID.
/// 4. Finally, it checks for an fd component followed by a numeric file descriptor.
/// Note, This function does not allocate.
/// Note, use negated PID if you only need validation and not the exact FD value.
/// SAFETY: This function returns Err(Errno::ELOOP) if PID is present but not equal to given PID.
#[allow(clippy::cognitive_complexity)]
pub fn proc_fd<P: AsRef<Path>>(pid: Pid, path: P) -> Result<Option<RawFd>, Errno> {
    let pid = pid.as_raw();
    #[allow(clippy::arithmetic_side_effects)]
    let (pid, need_fd) = if pid < 0 { (-pid, false) } else { (pid, true) };

    let path = path.as_ref().as_os_str().as_bytes();
    if !path.starts_with(PROC) {
        return Ok(None);
    }

    let path_without_proc = &path[PROC.len()..];
    let next_slash_index = path_without_proc
        .iter()
        .position(|&b| b == b'/')
        .unwrap_or(path_without_proc.len());
    let pid_section = &path_without_proc[..next_slash_index];
    if pid_section.is_empty() || !pid_section[0].is_ascii_digit() {
        return Ok(None); // Skip /proc/not-a-pid paths.
    }

    let after_pid_section = &path_without_proc[next_slash_index..];
    let start_of_interesting_part = if after_pid_section.starts_with(TASK) {
        let after_task_section = &after_pid_section[TASK.len()..];
        after_task_section
            .iter()
            .position(|&b| b == b'/')
            .map_or(after_task_section.len(), |p| p.saturating_add(TASK.len()))
    } else {
        0
    };

    let remaining_path = &after_pid_section[start_of_interesting_part..];
    if remaining_path.starts_with(FD) {
        // Check for valid file descriptor number after /fd/
        let fd_section = &remaining_path[FD.len()..];
        if fd_section.is_empty() {
            // Path ends with /fd, indicating it's a directory listing request, not an FD.
            return Ok(None);
        }

        if bytes_to_pid(pid_section) != pid {
            return Err(Errno::ELOOP); // PID mismatch detected.
        }

        let fd = if need_fd { bytes_to_fd(fd_section) } else { 0 };
        return Ok(Some(fd));
    } else if matches!(remaining_path, b"/cwd" | b"/exe" | b"/root") {
        // Validate PID only for special paths to avoid rejecting legitimate queries to other PIDs.
        if bytes_to_pid(pid_section) == pid {
            return Ok(Some(nix::libc::AT_FDCWD));
        } else {
            return Err(Errno::ELOOP);
        }
    }

    // No specific handling required, PID validation not necessary for paths not ending with /fd or special.
    Ok(None)
}

/// Fetches the environment variables:
/// - `LD_AUDIT`
/// - `LD_LIBRARY_PATH`
/// - `LD_PRELOAD`
/// for the specified process identified by `pid`.
///
/// # Arguments
/// * `pid` - The process identifier from which the environment variables should be retrieved.
///
/// # Returns
/// A result which is either:
/// * `Ok([Option<String>; 3])` - containing the paths if found and successfully parsed.
/// * `Err(VarError)` - In case of any error accessing the environment data.
///
/// # Errors
/// This function will return an error if:
/// - The process does not exist.
/// - The environment file is not accessible.
pub fn proc_ld_env(pid: Pid) -> Result<[Option<String>; 3], VarError> {
    // Construct the path to the environ file of the process.
    let mut buf = itoa::Buffer::new();
    let mut proc = PathBuf::from("/proc");
    proc.push(buf.format(pid.as_raw()));
    proc.push("environ");

    // Open the file.
    let mut file = match File::open(&proc) {
        Ok(file) => file,
        Err(_) => return Err(VarError::NotPresent),
    };

    // Read the contents of the file.
    let mut contents = Vec::new();
    if file.read_to_end(&mut contents).is_err() {
        return Err(VarError::NotPresent);
    }

    // Find the environment variables.
    // Items in /proc/pid/environ are split by null bytes.
    const NONE: std::option::Option<String> = None;
    let mut ret = [NONE; 3];
    for entry in contents.split(|&x| x == 0) {
        if entry.starts_with(b"LD_AUDIT=") {
            let value = &entry[b"LD_AUDIT=".len()..];
            let bytes = OsStr::from_bytes(value);
            ret[0] = Some(
                bytes
                    .to_str()
                    .map(String::from)
                    .ok_or_else(|| VarError::NotUnicode(bytes.to_os_string()))?,
            );
        } else if entry.starts_with(b"LD_PRELOAD=") {
            let value = &entry[b"LD_PRELOAD=".len()..];
            let bytes = OsStr::from_bytes(value);
            ret[1] = Some(
                bytes
                    .to_str()
                    .map(String::from)
                    .ok_or_else(|| VarError::NotUnicode(bytes.to_os_string()))?,
            );
        } else if entry.starts_with(b"LD_LIBRARY_PATH=") {
            let value = &entry[b"LD_LIBRARY_PATH=".len()..];
            let bytes = OsStr::from_bytes(value);
            ret[2] = Some(
                bytes
                    .to_str()
                    .map(String::from)
                    .ok_or_else(|| VarError::NotUnicode(bytes.to_os_string()))?,
            );
        }
    }

    Ok(ret)
}

#[inline]
fn bytes_to_pid(bytes: &[u8]) -> pid_t {
    bytes.iter().fold(0, |acc, &b| {
        acc.saturating_mul(10)
            .saturating_add(pid_t::from(b.saturating_sub(b'0')))
    })
}

#[inline]
fn bytes_to_fd(bytes: &[u8]) -> RawFd {
    bytes.iter().fold(0, |acc, &b| {
        acc.saturating_mul(10)
            .saturating_add(RawFd::from(b.saturating_sub(b'0')))
    })
}

/// Checks whether process memory usage is within the give maximum.
///
/// This function uses the `procfs` crate to obtain detailed memory maps
/// from `/proc/[pid]/smaps`. It sums multiple memory usage values reported in these maps
/// to calculate a more comprehensive total memory usage.
///
/// # Arguments
///
/// * `process` - `Process` instance representing the process.
///
/// # Returns
///
/// This function returns a `Result<bool, Errno>`. It returns Ok(true)
/// if the limit was exceeded Ok(false) otherwise. On failure, it
/// returns `Errno`.
///
/// # Errors
///
/// This function returns an error if it fails to retrieve the process's memory maps,
/// typically due to insufficient permissions or an invalid process ID.
pub fn proc_mem_limit(process: &Process, max: u64) -> Result<bool, Errno> {
    match process.smaps() {
        Ok(maps) => {
            let mut total_size: u64 = 0;
            for map in &maps.memory_maps {
                match &map.pathname {
                    MMapPath::Path(_)
                    | MMapPath::Anonymous
                    | MMapPath::Stack
                    | MMapPath::Other(_) => {
                        let pss = map.extension.map.get("Pss").copied().unwrap_or(0);
                        let private_dirty =
                            map.extension.map.get("Private_Dirty").copied().unwrap_or(0);
                        let shared_dirty =
                            map.extension.map.get("Shared_Dirty").copied().unwrap_or(0);

                        total_size = total_size.saturating_add(
                            pss.saturating_add(private_dirty)
                                .saturating_add(shared_dirty),
                        );

                        // Stop processing if total size exceeds or equals max
                        if total_size >= max {
                            return Ok(true);
                        }
                    }
                    _ => (),
                }
            }

            // If we're at this point, we did not hit the limit.
            Ok(false)
        }
        Err(_) => Err(Errno::last()),
    }
}

#[cfg(test)]
mod tests {
    use nix::sys::stat::umask;

    use super::*;

    #[test]
    fn test_invalid_pid() {
        let result = proc_umask(Pid::from_raw(i32::MAX));
        assert!(result.is_err(), "{result:?}");
    }

    #[test]
    fn test_parsing_valid_umask_values() {
        // This test sets various umask values and then checks if our function correctly identifies them.
        let umasks = [
            Mode::from_bits_truncate(0o0000),
            Mode::from_bits_truncate(0o0002),
            Mode::from_bits_truncate(0o0022),
            Mode::from_bits_truncate(0o0077),
            Mode::from_bits_truncate(0o0777),
        ];

        for &my_umask in &umasks {
            umask(my_umask);
            let result = proc_umask(Pid::this()).unwrap();
            assert_eq!(result, my_umask, "{result:o} != {my_umask:o}");
        }

        // Resetting the umask to a default value after test
        umask(Mode::from_bits_truncate(0o0022));
    }

    #[test]
    fn test_proc_fd() {
        let this = Pid::this();
        let that = Pid::from_raw(1);

        assert_eq!(
            proc_fd(this, format!("/proc/{this}/cwd")),
            Ok(Some(nix::libc::AT_FDCWD))
        );
        assert_eq!(
            proc_fd(this, format!("/proc/{this}/exe")),
            Ok(Some(nix::libc::AT_FDCWD))
        );
        assert_eq!(
            proc_fd(this, format!("/proc/{this}/root")),
            Ok(Some(nix::libc::AT_FDCWD))
        );
        assert_eq!(proc_fd(this, format!("/proc/{that}/fd")), Ok(None));
        assert_eq!(proc_fd(this, format!("/proc/{this}/fd")), Ok(None));
        assert_eq!(proc_fd(this, format!("/proc/{this}/fd/0")), Ok(Some(0)));
        assert_eq!(proc_fd(this, format!("/proc/{this}/fd/42")), Ok(Some(42)));
        assert_eq!(
            proc_fd(this, format!("/proc/{this}/fd/1984")),
            Ok(Some(1984))
        );
        assert_eq!(
            proc_fd(this, format!("/proc/{this}/task/{that}/fd/7")),
            Ok(Some(7))
        );

        assert_eq!(
            proc_fd(this, format!("/proc/{that}/cwd")),
            Err(Errno::ELOOP)
        );
        assert_eq!(
            proc_fd(this, format!("/proc/{that}/exe")),
            Err(Errno::ELOOP)
        );
        assert_eq!(
            proc_fd(this, format!("/proc/{that}/root")),
            Err(Errno::ELOOP)
        );
        assert_eq!(
            proc_fd(this, format!("/proc/{that}/fd/0")),
            Err(Errno::ELOOP)
        );
        assert_eq!(
            proc_fd(this, format!("/proc/{that}/task/{this}/fd/7")),
            Err(Errno::ELOOP)
        );
    }

    #[test]
    fn test_proc_ld_env() {
        let pid = Pid::this();
        let exp = std::env::var("LD_LIBRARY_PATH").unwrap_or_default();

        let env = proc_ld_env(pid);
        assert!(env.is_ok(), "env:{env:?}");
        let env = env.unwrap();

        assert_eq!(env[2], Some(exp), "env:{env:?}");
    }
}
