//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/syd-hex.rs: Hexadecimal encode/decode standard input.
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    env,
    fs::File,
    io::{BufRead, BufReader, Write},
    process::ExitCode,
};

use getargs::{Opt, Options};

fn main() -> ExitCode {
    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let args = env::args().skip(1).collect::<Vec<_>>();
    let mut opts = Options::new(args.iter().map(String::as_str));
    let mut encode = true;

    // SAFETY: We panic on parse errors.
    #[allow(clippy::disallowed_methods)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return ExitCode::SUCCESS;
            }
            Opt::Short('d') => {
                encode = false;
            }
            Opt::Short('e') => {
                encode = true;
            }
            _ => {
                eprintln!("Unknown option: {opt:?}!");
                return ExitCode::FAILURE;
            }
        }
    }

    // Lock stdin for efficient reading
    #[allow(clippy::disallowed_methods)]
    let mut input: Box<dyn BufRead> = match opts.positionals().next() {
        None | Some("-") => Box::new(std::io::stdin().lock()),
        Some(path) => Box::new(BufReader::new(File::open(path).expect("open file"))),
    };

    // Process the input in chunks to avoid loading it all into memory at once
    let mut buffer = [0; 64 * 1024]; // 64KB buffer size.
    while let Ok(count) = input.read(&mut buffer[..]) {
        if count == 0 {
            break; // End of input
        }
        if encode {
            // Hex-encode and write the chunk
            let encoded = const_hex::encode(&buffer[..count]);
            print!("{encoded}");
        } else {
            // Hex-decode and write the chunk
            #[allow(clippy::disallowed_methods)]
            match const_hex::decode(&buffer[..count]) {
                Ok(decoded) => std::io::stdout().write_all(&decoded).expect("write output"),
                Err(error) => {
                    eprintln!("Error decoding hex: {error}");
                    return ExitCode::FAILURE;
                }
            }
        }
    }

    ExitCode::SUCCESS
}

fn help() {
    println!("Usage: syd-hex [-hde] <file|->");
    println!("Given a file, hex-encode and print.");
    println!("Given no positional arguments, hex-encode standard input.");
    println!("Use -d to hex-decode rather than hex-encode.");
}
