use std::{
    cmp::Ordering,
    ffi::CString,
    hint::unreachable_unchecked,
    os::{
        fd::{AsRawFd, FromRawFd, OwnedFd},
        unix::io::RawFd,
    },
    ptr,
};

use libseccomp::ScmpFilterContext;
use nix::{
    errno::Errno,
    libc::c_char,
    sys::wait::{waitid, Id, WaitPidFlag, WaitStatus},
    unistd::{close, read, setpgid, write, Pid},
};

use crate::unshare::{child, config::Config, Child, Command, Executable};

type ChildPreExecFunc = Box<dyn Fn() -> Result<(), Errno>>;
type PipePair = ((RawFd, RawFd), (RawFd, RawFd));

#[derive(Debug)]
pub enum Exe<'a> {
    Library(&'a libloading::os::unix::Library),
    Program((*const c_char, Vec<*const c_char>)),
}

pub struct ChildInfo<'a> {
    pub exe: Exe<'a>,
    pub cfg: Config,
    pub pre_exec: Option<ChildPreExecFunc>,
    pub seccomp_filter: Option<ScmpFilterContext>,
    pub seccomp_pipefd: PipePair,
}

impl<'a> Drop for ChildInfo<'a> {
    fn drop(&mut self) {
        let _ = close(self.seccomp_pipefd.0 .0);
        let _ = close(self.seccomp_pipefd.0 .1);
        let _ = close(self.seccomp_pipefd.1 .0);
        let _ = close(self.seccomp_pipefd.1 .1);
    }
}

fn raw_with_null(arr: &Vec<CString>) -> Vec<*const c_char> {
    let mut vec = Vec::with_capacity(arr.len().saturating_add(1));
    for i in arr {
        vec.push(i.as_ptr());
    }
    vec.push(ptr::null());
    vec
}

impl Command {
    /// Spawn the command and return a handle that can be waited for
    pub fn spawn(mut self) -> Result<Child, Errno> {
        let exe = match self.exe {
            Executable::Library(ref lib) => Exe::Library(lib),
            Executable::Program((ref filename, ref args)) => {
                let c_args = raw_with_null(args);
                Exe::Program((filename.as_ptr(), c_args))
            }
        };

        let mut pid_fd: nix::libc::c_int = -1;
        let clone_flags = nix::libc::SIGCHLD | nix::libc::CLONE_PIDFD;
        let mut nstack = [0u8; 8 * 1024];
        let child_info = Box::new(ChildInfo {
            exe,
            cfg: self.config,
            pre_exec: std::mem::take(&mut self.pre_exec),
            seccomp_filter: std::mem::take(&mut self.seccomp_filter),
            seccomp_pipefd: self.seccomp_pipefd,
        });
        let child_info_ptr: *mut nix::libc::c_void =
            Box::into_raw(child_info) as *mut nix::libc::c_void;

        // SAFETY: In libc we trust.
        let child = unsafe {
            nix::libc::clone(
                child::child_after_clone
                    as extern "C" fn(*mut nix::libc::c_void) -> nix::libc::c_int,
                nstack.last_mut().unwrap_unchecked() as *mut u8 as *mut nix::libc::c_void,
                clone_flags,
                child_info_ptr,
                &mut pid_fd,
                std::ptr::null_mut::<nix::libc::c_int>(),
                std::ptr::null_mut::<nix::libc::c_int>(),
            )
        };

        match child.cmp(&0) {
            Ordering::Less => {
                // SAFETY: Reconstruct and drop.
                let _ = unsafe { Box::from_raw(child_info_ptr as *mut ChildInfo) };
                Err(Errno::last())
            }
            Ordering::Greater => {
                let child = Pid::from_raw(child);

                let seccomp_fd = match self.after_start(child, pid_fd) {
                    Ok(seccomp_fd) => seccomp_fd,
                    Err(e) => loop {
                        match waitid(Id::PIDFd(pid_fd), WaitPidFlag::WEXITED) {
                            Ok(WaitStatus::Exited(_, errno)) => return Err(Errno::from_i32(errno)),
                            Err(Errno::EINTR) => {}
                            _ => return Err(e),
                        }
                    },
                };

                Ok(Child {
                    pid: child.into(),
                    pid_fd,
                    seccomp_fd,
                    status: None,
                })
            }
            // SAFETY: This can never happen because clone child
            // jumps to the specified function.
            _ => unsafe { unreachable_unchecked() },
        }
    }

    #[allow(clippy::cognitive_complexity)]
    fn after_start(mut self, pid: Pid, pid_fd: RawFd) -> Result<RawFd, Errno> {
        if self.config.make_group_leader {
            setpgid(pid, pid)?;
        }

        if let Some(ref mut callback) = self.before_unfreeze {
            #[allow(clippy::cast_sign_loss)]
            callback(i32::from(pid) as u32)?;
        }

        // Close the read end of the first pipe
        // Close the write end of the second pipe
        let _ = close(self.seccomp_pipefd.0 .0);
        let _ = close(self.seccomp_pipefd.1 .1);
        // SAFETY: seccomp_pipefds hold return value of pipe2 which are valid FDs.
        let pipe_ro = unsafe { OwnedFd::from_raw_fd(self.seccomp_pipefd.1 .0) };
        // SAFETY: seccomp_pipefds hold return value of pipe2 which are valid FDs.
        let pipe_rw = unsafe { OwnedFd::from_raw_fd(self.seccomp_pipefd.0 .1) };

        // Read the value of the file descriptor from the pipe.
        let mut buf = vec![0u8; std::mem::size_of::<RawFd>()];
        read(pipe_ro.as_raw_fd(), &mut buf)?;
        let fd = match buf.as_slice().try_into() {
            Ok(bytes) => RawFd::from_le_bytes(bytes),
            Err(_) => return Err(Errno::EINVAL),
        };

        // SAFETY: In libc we trust.
        let fd = match unsafe { nix::libc::syscall(nix::libc::SYS_pidfd_getfd, pid_fd, fd, 0) } {
            e if e < 0 => return Err(Errno::last()),
            fd => fd as RawFd,
        };

        // Send success notification to the child.
        let ack = [7u8];
        if let Err(errno) = write(pipe_rw.as_raw_fd(), &ack) {
            let _ = close(fd);
            return Err(errno);
        }

        Ok(fd)
    }
}
